from collections import OrderedDict
from dataclasses import asdict
from typing import Literal, Optional, Union

import numpy as np
import torch
import torch.nn as nn
from gymnasium import spaces
from torch._dynamo.eval_frame import OptimizedModule

from agilerl.modules import EvolvableModule, ModuleDict
from agilerl.modules.configs import (
    CnnNetConfig,
    LstmNetConfig,
    MlpNetConfig,
    MultiInputNetConfig,
    NetConfig,
    SimBaNetConfig,
)
from agilerl.modules.custom_components import (
    GumbelSoftmax,
    NewGELU,
    NoisyLinear,
    ResidualBlock,
    SimbaResidualBlock,
)
from agilerl.typing import DeviceType, NetConfigType

TupleorInt = Union[tuple[int, ...], int]


def compile_model(
    model: Union[nn.Module, ModuleDict[EvolvableModule]],
    mode: Optional[str] = "default",
) -> Union[OptimizedModule, ModuleDict[EvolvableModule]]:
    """Compiles torch model if not already compiled

    :param model: torch model
    :type model: nn.Module | ModuleDict[EvolvableModule]
    :param mode: torch compile mode, defaults to "default"
    :type mode: str, optional
    :return: compiled model
    :rtype: OptimizedModule | ModuleDict[OptimizedModule]
    """
    if isinstance(model, ModuleDict):
        compiled_model = ModuleDict(
            {
                agent_id: compile_model(module, mode)
                for agent_id, module in model.items()
            }
        )
        compiled_model.last_mutation_attr = model.last_mutation_attr
        return compiled_model

    if not isinstance(model, OptimizedModule) and mode is not None:
        compiled_model = torch.compile(model, mode=mode, dynamic=True)
    else:
        compiled_model = model

    return compiled_model


def is_image_space(space: spaces.Space) -> bool:
    """Check if the space is an image space. We ignore dtype and number of channels
    checks.

    :param space: Input space
    :type space: spaces.Space

    :return: True if the space is an image space, False otherwise
    :rtype: bool
    """
    return isinstance(space, spaces.Box) and len(space.shape) == 3


def is_box_space_ndim(space: spaces.Space, ndim: int) -> bool:
    """Check if the space is a Box space with the given number of dimensions.

    :param space: Input space
    :type space: spaces.Space
    :param ndim: Number of dimensions
    :type ndim: int

    :return: True if the space is a Box space with the given number of dimensions, False otherwise
    """
    return isinstance(space, spaces.Box) and len(space.shape) == ndim


def is_vector_space(space: spaces.Space) -> bool:
    """Check if the space is a vector space.

    :param space: Input space
    :type space: spaces.Space

    :return: True if the space is a vector space, False otherwise
    :rtype: bool
    """
    return (
        (isinstance(space, spaces.Box) and len(space.shape) in [0, 1])
        or isinstance(space, spaces.Discrete)
        or isinstance(space, spaces.MultiDiscrete)
    )


def config_from_dict(config_dict: NetConfigType) -> NetConfig:
    """Get the class of the net config from the dictionary.

    :param config_dict: The dictionary to get the class from.
    :type config_dict: NetConfigType
    :return: The net config class.
    :rtype: NetConfig
    """
    config_keys = config_dict.keys()
    if "hidden_size" in config_keys:
        if "num_layers" in config_keys:
            config_cls = LstmNetConfig
        elif "num_blocks" in config_keys:
            config_cls = SimBaNetConfig
        else:
            config_cls = MlpNetConfig
    elif "channel_size" in config_keys:
        config_cls = CnnNetConfig
    elif any(
        key in MultiInputNetConfig.__dataclass_fields__.keys() for key in config_keys
    ):
        config_cls = MultiInputNetConfig
    else:
        raise ValueError(
            f"Unable to determine net config class from: {config_dict}. "
            "Please verify that the keys correspond to the arguments of the net config class."
        )

    return config_cls.from_dict(config_dict)


def tuple_to_dict_space(tuple_space: spaces.Tuple) -> spaces.Dict:
    """Converts a Tuple observation space to a Dict observation space.

    :param tuple_space: Tuple observation space
    :type tuple_space: spaces.Tuple
    :return: Dictionary observation space
    :rtype: spaces.Dict
    """
    return spaces.Dict({str(i): space for i, space in enumerate(tuple_space.spaces)})


def tuple_to_dict_obs(tuple_obs: tuple) -> dict:
    """Converts a tuple observation to a Python dictionary

    :param tuple_obs: Tuple observation
    :type tuple_obs: tuple
    :return: Dictionary observation
    :rtype: dict
    """
    return {str(i): obs for i, obs in enumerate(tuple_obs)}


def get_default_encoder_config(
    observation_space: spaces.Space, simba: bool = False, recurrent: bool = False
) -> NetConfigType:
    """Get the default configuration for the encoder network based on the observation space.

    :param observation_space: Observation space of the environment.
    :type observation_space: spaces.Space
    :param simba: Whether to use SimBA encoder.
    :type simba: bool
    :param recurrent: Whether to use recurrent encoder.
    :type recurrent: bool

    :return: Default configuration for the encoder network.
    :rtype: dict[str, Any]
    """
    if isinstance(observation_space, (spaces.Dict, spaces.Tuple)):
        config = MultiInputNetConfig(output_activation=None)
    elif is_image_space(observation_space):
        config = CnnNetConfig(
            channel_size=[32, 32],
            kernel_size=[3, 3],
            stride_size=[1, 1],
            output_activation=None,
        )
    elif simba:
        config = SimBaNetConfig(hidden_size=128, num_blocks=2, output_activation=None)
    elif recurrent:
        config = LstmNetConfig(
            hidden_state_size=128, num_layers=2, output_activation=None
        )
    else:
        config = MlpNetConfig(
            hidden_size=[64, 64], output_activation=None, output_vanish=False
        )

    return asdict(config)


def contains_moduledict(module: nn.Module) -> bool:
    """Check if a module contains a ModuleDict.

    :param module: Input module
    :type module: nn.Module

    :return: True if module contains a ModuleDict, False otherwise
    :rtype: bool
    """
    for submodule in module.modules():
        if isinstance(submodule, nn.ModuleDict):
            return True
    return False


def get_module_dict(module: nn.Module) -> nn.ModuleDict:
    """Get the ModuleDict from a module.

    :param module: Input module
    :type module: nn.Module

    :return: ModuleDict from module
    :rtype: dict[str, nn.Module]
    """
    for submodule in module.modules():
        if isinstance(submodule, nn.ModuleDict):
            return submodule
    return None


def get_batch_norm_layer(
    name: str, num_features: int, device: DeviceType = "cpu"
) -> nn.Module:
    """Return batch normalization layer for corresponding batch normalization name.

    :param name: Batch normalization layer name
    :type name: str
    :param layer_size: The layer after which the batch normalization layer will be applied
    :type layer_size: int

    :return: Batch normalization layer
    :rtype: nn.Module
    """
    batch_norm_layers = {
        "1d": nn.BatchNorm1d,
        "2d": nn.BatchNorm2d,
        "3d": nn.BatchNorm3d,
    }

    return batch_norm_layers[name](num_features, device=device)


def get_conv_layer(
    conv_layer_name: Literal["Conv2d", "Conv3d"],
    in_channels: int,
    out_channels: int,
    kernel_size: TupleorInt,
    stride: TupleorInt = 1,
    padding: TupleorInt = 0,
    device: DeviceType = "cpu",
) -> nn.Module:
    """Return convolutional layer for corresponding convolutional layer name.

    :param conv_layer_name: Convolutional layer name
    :type conv_layer_name: str
    :param in_channels: Number of input channels to convolutional layer
    :type in_channels: int
    :param out_channels: Number of output channels from convolutional layer
    :type out_channels: int
    :param kernel_size: Kernel size of convolutional layer
    :type kernel_size: int or tuple[int]
    :param stride: Stride size of convolutional layer
    :type stride: int or tuple[int]
    :param padding: Convolutional layer padding
    :type padding: int or tuple[int]

    :return: Convolutional layer
    :rtype: nn.Module
    """
    if conv_layer_name not in ["Conv1d", "Conv2d", "Conv3d"]:
        raise ValueError(
            f"Invalid convolutional layer {conv_layer_name}. Must be one of 'Conv1d', 'Conv2d', 'Conv3d'."
        )

    convolutional_layers = {
        "1d": nn.Conv1d,
        "2d": nn.Conv2d,
        "3d": nn.Conv3d,
    }

    # remove 'Conv' from the name if it is present
    conv_layer_name = conv_layer_name.replace("Conv", "")
    return convolutional_layers[conv_layer_name](
        in_channels, out_channels, kernel_size, stride, padding, device=device
    )


def get_normalization(
    normalization_name: str, layer_size: int, device: DeviceType = "cpu"
) -> nn.Module:
    """Returns normalization layer for corresponding normalization name.

    :param normalization_names: Normalization layer name
    :type normalization_names: str
    :param layer_size: The layer after which the normalization layer will be applied
    :param layer_size: int

    :return: Normalization layer
    :rtype: nn.Module
    """
    normalization_functions = {
        "BatchNorm2d": nn.BatchNorm2d,
        "BatchNorm3d": nn.BatchNorm3d,
        "InstanceNorm2d": nn.InstanceNorm2d,
        "InstanceNorm3d": nn.InstanceNorm3d,
        "LayerNorm": nn.LayerNorm,
    }

    return normalization_functions[normalization_name](layer_size, device=device)


def get_activation(activation_name: Optional[str], new_gelu: bool = False) -> nn.Module:
    """Returns activation function for corresponding activation name.

    :param activation_names: Activation function name
    :type activation_names: str
    """
    activation_functions = {
        "Tanh": nn.Tanh,
        "ReLU": nn.ReLU,
        "ELU": nn.ELU,
        "Softsign": nn.Softsign,
        "Sigmoid": nn.Sigmoid,
        "GumbelSoftmax": GumbelSoftmax,
        "Softplus": nn.Softplus,
        "Softmax": nn.Softmax,
        "LeakyReLU": nn.LeakyReLU,
        "PReLU": nn.PReLU,
        "GELU": nn.GELU if not new_gelu else NewGELU,
        "Identity": nn.Identity,
    }

    activation_name = activation_name if activation_name is not None else "Identity"
    return (
        activation_functions[activation_name](dim=-1)
        if activation_name == "Softmax"
        else activation_functions[activation_name]()
    )


def get_pooling(
    pooling_name: str,
    kernel_size: Union[tuple[int, ...], int],
    stride: Union[tuple[int, ...], int],
    padding: Union[tuple[int, ...], int],
) -> nn.Module:
    """Returns pooling layer for corresponding activation name.

    :param pooling_names: Pooling layer name
    :type pooling_names: str
    :param kernel_size: Pooling layer kernel size
    :type kernel_size: int or tuple[int]
    :param stride: Pooling layer stride
    :type stride: int or tuple[int]
    :param padding: Pooling layer padding
    :type padding: int or tuple[int]

    :return: Pooling layer
    :rtype: nn.Module
    """
    pooling_functions = {
        "MaxPool2d": nn.MaxPool2d,
        "MaxPool3d": nn.MaxPool3d,
        "AvgPool2d": nn.AvgPool2d,
        "AvgPool3d": nn.AvgPool3d,
    }

    return pooling_functions[pooling_name](kernel_size, stride, padding)


LayerType = Union[nn.Module, GumbelSoftmax, NoisyLinear]


def layer_init(
    layer: LayerType, std: float = np.sqrt(2), bias_const: float = 0.0
) -> nn.Module:
    """
    Initializes the weights and biases of a layer.

    :param layer: The layer to initialize.
    :type layer: nn.Module
    :param std: The standard deviation for weight initialization. Defaults to np.sqrt(2).
    :type std: float, optional
    :param bias_const: The constant value for bias initialization. Defaults to 0.0.
    :type bias_const: float, optional

    :return: The initialized layer.
    :rtype: nn.Module
    """
    if hasattr(layer, "weight"):
        torch.nn.init.orthogonal_(layer.weight, std)

    elif hasattr(layer, "weight_mu") and hasattr(layer, "weight_sigma"):
        torch.nn.init.orthogonal_(layer.weight_mu, std)
        torch.nn.init.orthogonal_(layer.weight_sigma, std)

    if hasattr(layer, "bias"):
        torch.nn.init.constant_(layer.bias, bias_const)

    elif hasattr(layer, "bias_mu") and hasattr(layer, "bias_sigma"):
        torch.nn.init.constant_(layer.bias_mu, bias_const)
        torch.nn.init.constant_(layer.bias_sigma, bias_const)

    return layer


def init_weights_gaussian(m: nn.Module, mean: float, std: float) -> None:
    """Initialize weights of a module using Gaussian distribution.

    :param m: Module to initialize
    :type m: nn.Module
    :param mean: Mean of the Gaussian distribution
    :type mean: float
    :param std: Standard deviation of the Gaussian distribution
    :type std: float
    """
    if isinstance(m, nn.Linear):
        torch.nn.init.normal_(m.weight, mean=mean, std=std)
        if m.bias is not None:
            torch.nn.init.constant_(m.bias, 0)


def create_cnn(
    block_type: Literal["Conv2d", "Conv3d"],
    in_channels: int,
    channel_size: list[int],
    kernel_size: list[TupleorInt],
    stride_size: list[TupleorInt],
    name: str = "cnn",
    init_layers: bool = True,
    layer_norm: bool = False,
    activation_fn: str = "ReLU",
    device: DeviceType = "cpu",
) -> dict[str, nn.Module]:
    """
    Build a convolutional block.

    :param block_type: Type of convolutional block.
    :type block_type: Literal["Conv2d", "Conv3d"]
    :param in_channels: Number of input channels.
    :type in_channels: int
    :param channel_size: List of channel sizes for each layer.
    :type channel_size: list[int]
    :param kernel_size: List of kernel sizes for each layer.
    :type kernel_size: list[int]
    :param stride_size: List of stride sizes for each layer.
    :type stride_size: list[int]
    :param name: Name of the block.
    :type name: str
    :param init_layers: Whether to initialize the layers. Defaults to True.
    :type init_layers: bool, optional
    :param layer_norm: Whether to use layer normalization. Defaults to False.
    :type layer_norm: bool, optional
    :param activation_fn: Activation function to use. Defaults to "ReLU".
    :type activation_fn: str, optional
    :param device: Device to use. Defaults to "cpu".
    :type device: DeviceType, optional

    :return: Convolutional block.
    :rtype: dict[str, nn.Module]
    """
    net_dict = OrderedDict()
    channel_size = [in_channels] + channel_size
    for l_no in range(1, len(channel_size)):
        net_dict[f"{name}_conv_layer_{str(l_no)}"] = get_conv_layer(
            conv_layer_name=block_type,
            in_channels=channel_size[l_no - 1],
            out_channels=channel_size[l_no],
            kernel_size=kernel_size[l_no - 1],
            stride=stride_size[l_no - 1],
            device=device,
        )
        if init_layers:
            net_dict[f"{name}_conv_layer_{str(l_no)}"] = layer_init(
                net_dict[f"{name}_conv_layer_{str(l_no)}"]
            )
        if layer_norm:
            net_dict[f"{name}_layer_norm_{str(l_no)}"] = get_batch_norm_layer(
                block_type.replace("Conv", ""),
                num_features=channel_size[l_no],
                device=device,
            )
        net_dict[f"{name}_activation_{str(l_no)}"] = get_activation(activation_fn)

    return net_dict


MlpLayer = Union[nn.Linear, NoisyLinear, nn.LayerNorm]


def create_mlp(
    input_size: int,
    output_size: int,
    hidden_size: list[int],
    output_vanish: bool,
    output_activation: Optional[str] = None,
    noisy: bool = False,
    init_layers: bool = True,
    layer_norm: bool = False,
    output_layernorm: bool = False,
    activation: str = "ReLU",
    noise_std: float = 0.1,
    device: DeviceType = "cpu",
    new_gelu: bool = False,
    name: str = "mlp",
) -> nn.Sequential:
    """Creates and returns multi-layer perceptron.

    :param input_size: Number of input features.
    :type input_size: int
    :param output_size: Number of output features.
    :type output_size: int
    :param hidden_size: List of hidden layer sizes.
    :type hidden_size: list[int]
    :param output_vanish: Whether to initialize output layer weights to a small value.
    :type output_vanish: bool
    :param output_activation: Activation function for output layer.
    :type output_activation: Optional[str]
    :param noisy: Whether to use noisy layers.
    :type noisy: bool, optional
    :param init_layers: Whether to initialize the layers.
    :type init_layers: bool, optional
    :param layer_norm: Whether to use layer normalization.
    :type layer_norm: bool, optional
    :param output_layernorm: Whether to use layer normalization for the output layer.
    :type output_layernorm: bool, optional
    :param activation: Activation function for hidden layers.
    :type activation: str, optional
    :param noise_std: Standard deviation of noise for noisy layers.
    :type noise_std: float, optional
    :param name: Name of the network.
    :type name: str, default "mlp"

    :return: Multi-layer perceptron.
    :rtype: nn.Sequential
    """
    net_dict: dict[str, MlpLayer] = OrderedDict()
    hidden_size = [input_size] + hidden_size
    for l_no in range(1, len(hidden_size)):
        if noisy:  # Add linear layer
            net_dict[f"{name}_linear_layer_{str(l_no)}"] = NoisyLinear(
                hidden_size[l_no - 1], hidden_size[l_no], noise_std, device=device
            )
        else:
            net_dict[f"{name}_linear_layer_{str(l_no)}"] = nn.Linear(
                hidden_size[l_no - 1], hidden_size[l_no], device=device
            )

        if init_layers:  # Initialize layer weights
            net_dict[f"{name}_linear_layer_{str(l_no)}"] = layer_init(
                net_dict[f"{name}_linear_layer_{str(l_no)}"]
            )

        if layer_norm:  # Add layer normalization
            net_dict[f"{name}_layer_norm_{str(l_no)}"] = nn.LayerNorm(
                hidden_size[l_no], device=device
            )

        # Add activation function
        net_dict[f"{name}_activation_{str(l_no)}"] = get_activation(
            activation, new_gelu
        )

    # Output layer
    if noisy:
        output_layer = NoisyLinear(
            hidden_size[-1], output_size, noise_std, device=device
        )
    else:
        output_layer = nn.Linear(hidden_size[-1], output_size, device=device)

    if init_layers:
        output_layer = layer_init(output_layer)

    if output_vanish:
        if noisy:
            output_layer.weight_mu.data.mul_(0.1)
            output_layer.bias_mu.data.mul_(0.1)
            output_layer.weight_sigma.data.mul_(0.1)
            output_layer.bias_sigma.data.mul_(0.1)
        else:
            output_layer.weight.data.mul_(0.1)
            output_layer.bias.data.mul_(0.1)

    net_dict[f"{name}_linear_layer_output"] = output_layer

    if output_layernorm:
        net_dict[f"{name}_layer_norm_output"] = nn.LayerNorm(
            output_size, device=device, elementwise_affine=False
        )

    net_dict[f"{name}_activation_output"] = get_activation(
        activation_name=output_activation, new_gelu=new_gelu
    )
    return nn.Sequential(net_dict)


def create_simba(
    input_size: int,
    output_size: int,
    hidden_size: int,
    num_blocks: int,
    output_activation: Optional[str] = None,
    scale_factor: float = 4.0,
    device: DeviceType = "cpu",
    name: str = "simba",
) -> nn.Sequential:
    """Creates a number of SimBa residual blocks.

    Paper: https://arxiv.org/abs/2410.09754.

    :param input_size: Number of input features.
    :type input_size: int
    :param output_size: Number of output features.
    :type output_size: int
    :param hidden_size: Number of hidden units.
    :type hidden_size: int
    :param num_blocks: Number of residual blocks.
    :type num_blocks: int
    :param output_activation: Activation function for output layer.
    :type output_activation: Optional[str]
    :param scale_factor: Scale factor for the hidden layer.
    :type scale_factor: float, optional
    :param device: Device to use. Defaults to "cpu".
    :type device: DeviceType, optional
    :param name: Name of the network.
    :type name: str, default "simba"

    :return: Residual block.
    :rtype: nn.Sequential
    """
    net_dict: dict[str, nn.Module] = OrderedDict()

    # Initial dense layer
    net_dict[f"{name}_linear_layer_input"] = nn.Linear(
        input_size, hidden_size, device=device
    )
    nn.init.orthogonal_(net_dict[f"{name}_linear_layer_input"].weight)
    for l_no in range(1, num_blocks + 1):
        net_dict[f"{name}_residual_block_{str(l_no)}"] = SimbaResidualBlock(
            hidden_size, scale_factor=scale_factor, device=device
        )

    # Final layer norm and output dense
    net_dict[f"{name}_layer_norm_output"] = nn.LayerNorm(hidden_size, device=device)
    net_dict[f"{name}_linear_layer_output"] = nn.Linear(
        hidden_size, output_size, device=device
    )
    nn.init.orthogonal_(net_dict[f"{name}_linear_layer_output"].weight)

    net_dict[f"{name}_activation_output"] = get_activation(
        activation_name=output_activation
    )

    return nn.Sequential(net_dict)


def create_resnet(
    input_channels: int,
    channel_size: int,
    kernel_size: int,
    stride_size: int,
    num_blocks: int,
    scale_factor: int = 4,
    device: str = "cpu",
    name: str = "resnet",
) -> dict[str, nn.Module]:
    """Creates a number of residual blocks for image-based inputs.

    Paper: https://arxiv.org/abs/1512.03385.

    :param input_channels: Number of input channels.
    :type input_channels: int
    :param channel_size: Number of output channels.
    :type channel_size: int
    :param kernel_size: Kernel size of the convolutional layer.
    :type kernel_size: int
    :param stride_size: Stride size of the convolutional layer.
    :type stride_size: int
    :param num_blocks: Number of residual blocks.
    :type num_blocks: int
    :param scale_factor: Scale factor for the hidden layer.
    :type scale_factor: int, optional
    :param device: Device to use. Defaults to "cpu".
    :type device: DeviceType, optional
    :param name: Name of the network.
    :type name: str, default "resnet"

    :return: Residual block.
    :rtype: nn.Sequential
    """
    net_dict = OrderedDict()

    # Initial convolutional layer
    net_dict[f"{name}_conv_input"] = nn.Conv2d(
        input_channels,
        channel_size,
        kernel_size=kernel_size,
        stride=stride_size,
        padding=(kernel_size - 1) // 2,
        bias=False,
        device=device,
    )
    nn.init.kaiming_uniform_(net_dict[f"{name}_conv_input"].weight)

    for l_no in range(1, num_blocks + 1):
        net_dict[f"{name}_residual_block_{l_no}"] = ResidualBlock(
            in_channels=channel_size,
            kernel_size=kernel_size,
            scale_factor=scale_factor,
            device=device,
        )

    return net_dict



import numpy as np
import pytest
import torch
from tensordict import TensorDict

from agilerl.components.replay_buffer import (
    MultiStepReplayBuffer,
    PrioritizedReplayBuffer,
    ReplayBuffer,
)
from agilerl.components.segment_tree import MinSegmentTree, SumSegmentTree


##### ReplayBuffer class tests #####
def test_create_instance_with_valid_arguments():
    """Test that a ReplayBuffer can be created with valid arguments."""
    max_size = 100
    device = "cpu"
    dtype = torch.float32

    buffer = ReplayBuffer(max_size, device, dtype)

    assert buffer.max_size == max_size
    assert buffer.device == device
    assert buffer.dtype == dtype
    assert buffer.counter == 0
    assert buffer.initialized is False
    assert buffer._cursor == 0
    assert buffer._size == 0
    assert buffer._storage is None


def test_get_length_of_memory():
    """Test that the length of the buffer can be retrieved with __len__ method."""
    max_size = 100
    buffer = ReplayBuffer(max_size)

    # Add experiences to memory
    data = TensorDict(
        {
            "state": torch.tensor([1, 2, 3]),
            "action": torch.tensor([0]),
            "reward": torch.tensor([1.0]),
        },
    )
    data = data.unsqueeze(0)
    data.batch_size = [1]
    buffer.add(data)

    data = TensorDict(
        {
            "state": torch.tensor([4, 5, 6]),
            "action": torch.tensor([1]),
            "reward": torch.tensor([2.0]),
        },
    )
    data = data.unsqueeze(0)
    data.batch_size = [1]
    buffer.add(data)

    assert len(buffer) == 2
    assert buffer.size == 2


def test_buffer_initialization():
    """Test that the buffer correctly initializes storage when adding first data."""
    buffer = ReplayBuffer(max_size=1000)

    # Create data
    data = TensorDict(
        {
            "state": torch.tensor([1, 2, 3]),
            "action": torch.tensor([0]),
            "reward": torch.tensor([1.0]),
        },
    )
    data = data.unsqueeze(0)
    data.batch_size = [1]

    # Add to buffer
    buffer.add(data)

    # Check initialization status
    assert buffer.initialized is True
    assert buffer._storage is not None
    assert len(buffer) == 1
    assert buffer.counter == 1


def test_add_experience_when_buffer_full():
    """Test that when buffer is full, old experiences are overwritten."""
    buffer = ReplayBuffer(max_size=2)

    # Add three experiences (more than buffer size)
    data1 = TensorDict(
        {
            "state": torch.tensor([1, 2, 3]),
            "action": torch.tensor([0]),
            "reward": torch.tensor([1.0]),
        },
    )
    data1 = data1.unsqueeze(0)
    data1.batch_size = [1]
    buffer.add(data1)

    data2 = TensorDict(
        {
            "state": torch.tensor([4, 5, 6]),
            "action": torch.tensor([1]),
            "reward": torch.tensor([2.0]),
        },
    )
    data2 = data2.unsqueeze(0)
    data2.batch_size = [1]
    buffer.add(data2)

    data3 = TensorDict(
        {
            "state": torch.tensor([7, 8, 9]),
            "action": torch.tensor([2]),
            "reward": torch.tensor([3.0]),
        },
    )
    data3 = data3.unsqueeze(0)
    data3.batch_size = [1]
    buffer.add(data3)

    # Check that buffer has max_size elements
    assert len(buffer) == 2
    assert buffer.is_full

    # Check that oldest experience was overwritten
    sample = buffer.sample(2)
    assert sample.keys() == data1.keys()

    # Either data2 and data3 should be in the buffer, data1 should be overwritten
    assert any(
        torch.all(sample["state"][i] == data2["state"])
        or torch.all(sample["state"][i] == data3["state"])
        for i in range(2)
    )


def test_add_vectorized_experiences():
    """Test adding vectorized experiences."""
    buffer = ReplayBuffer(max_size=10)

    # Create batched data
    batch_data = TensorDict(
        {
            "state": torch.tensor([[1, 2, 3], [4, 5, 6]]),
            "action": torch.tensor([[0], [1]]),
            "reward": torch.tensor([[1.0], [2.0]]),
        },
    )

    # Add batch to buffer
    batch_data.batch_size = [2]
    buffer.add(batch_data)

    # Check buffer state
    assert len(buffer) == 2
    assert buffer.counter == 2


def test_sample_experiences():
    """Test sampling experiences from the buffer."""
    buffer = ReplayBuffer(max_size=100)

    # Add multiple experiences
    for i in range(10):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Sample experiences
    batch_size = 5
    samples = buffer.sample(batch_size)

    # Verify sample structure
    assert isinstance(samples, TensorDict)
    assert samples.batch_size == torch.Size([batch_size])
    assert "state" in samples
    assert "action" in samples
    assert "reward" in samples
    assert samples["state"].shape == (batch_size, 3)
    assert samples["action"].shape == (batch_size, 1)
    assert samples["reward"].shape == (batch_size, 1)


def test_sample_experiences_with_images():
    """Test sampling experiences with image observations."""
    buffer = ReplayBuffer(max_size=100)

    # Add multiple experiences with image observations
    for i in range(10):
        # Create image-like tensor (C, H, W)
        img = torch.ones((3, 84, 84)) * i
        data = TensorDict(
            {
                "state": img,
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Sample experiences
    batch_size = 5
    samples = buffer.sample(batch_size)

    # Verify sample structure for image observations
    assert isinstance(samples, TensorDict)
    assert samples.batch_size == torch.Size([batch_size])
    assert "state" in samples
    assert samples["state"].shape == (batch_size, 3, 84, 84)
    assert samples["action"].shape == (batch_size, 1)
    assert samples["reward"].shape == (batch_size, 1)


def test_sample_experiences_with_dict_obs():
    """Test sampling experiences with dictionary observations."""
    buffer = ReplayBuffer(max_size=100)

    # Add multiple experiences with dictionary observations
    for i in range(10):
        # Create dictionary observation
        obs = TensorDict(
            {
                "image": torch.ones((3, 84, 84)) * i,
                "vector": torch.tensor([i, i + 1, i + 2]),
            },
        )
        data = TensorDict(
            {
                "state": obs,
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Sample experiences
    batch_size = 5
    samples = buffer.sample(batch_size)

    # Verify sample structure for dictionary observations
    assert isinstance(samples, TensorDict)
    assert samples.batch_size == torch.Size([batch_size])
    assert "state" in samples
    assert isinstance(samples["state"], TensorDict)
    assert "image" in samples["state"]
    assert "vector" in samples["state"]
    assert samples["state"]["image"].shape == (batch_size, 3, 84, 84)
    assert samples["state"]["vector"].shape == (batch_size, 3)
    assert samples["action"].shape == (batch_size, 1)
    assert samples["reward"].shape == (batch_size, 1)


def test_sample_with_indices():
    """Test sampling experiences with indices returned."""
    buffer = ReplayBuffer(max_size=100)

    # Add multiple experiences
    for i in range(10):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Sample experiences with indices
    batch_size = 5
    samples = buffer.sample(batch_size, return_idx=True)

    # Verify indices are returned
    assert "idxs" in samples
    assert samples["idxs"].shape == (batch_size,)
    assert all(0 <= idx < len(buffer) for idx in samples["idxs"])


def test_clear_buffer():
    """Test clearing the buffer."""
    buffer = ReplayBuffer(max_size=100)

    # Add experiences
    data = TensorDict(
        {
            "state": torch.tensor([1, 2, 3]),
            "action": torch.tensor([0]),
            "reward": torch.tensor([1.0]),
        },
    )
    data = data.unsqueeze(0)
    data.batch_size = [1]
    buffer.add(data)

    # Clear buffer
    buffer.clear()

    # Check buffer state
    assert len(buffer) == 0
    assert buffer._size == 0
    assert buffer._cursor == 0
    assert buffer._storage is None
    assert not buffer.initialized


##### MultiStepReplayBuffer class tests #####
def test_nstep_buffer_initialization():
    """Test initialization of MultiStepReplayBuffer."""
    max_size = 1000
    n_step = 3
    gamma = 0.99

    buffer = MultiStepReplayBuffer(max_size, n_step, gamma)

    assert buffer.max_size == max_size
    assert buffer.n_step == n_step
    assert buffer.gamma == gamma
    assert len(buffer.n_step_buffer) == 0


def test_nstep_buffer_add():
    """Test adding transitions to MultiStepReplayBuffer."""
    buffer = MultiStepReplayBuffer(max_size=1000, n_step=3, gamma=0.99)

    # Define transition fields required for n-step returns
    buffer.reward_key = "reward"
    buffer.done_key = "done"
    buffer.ns_key = "next_state"

    # Add a transition
    data = TensorDict(
        {
            "state": torch.tensor([1, 2, 3]),
            "action": torch.tensor([0]),
            "reward": torch.tensor([1.0]),
            "next_state": torch.tensor([4, 5, 6]),
            "done": torch.tensor([False]),
        },
    )
    data = data.unsqueeze(0)
    data.batch_size = [1]
    buffer.add(data)

    # Check that transition is in n-step buffer but not in main buffer yet
    assert len(buffer.n_step_buffer) == 1
    assert len(buffer) == 0

    # Add more transitions to complete n-step return
    for i in range(2):
        data = TensorDict(
            {
                "state": torch.tensor([i + 4, i + 5, i + 6]),
                "action": torch.tensor([i + 1]),
                "reward": torch.tensor([float(i + 2)]),
                "next_state": torch.tensor([i + 7, i + 8, i + 9]),
                "done": torch.tensor([False]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Now a transition should be in the main buffer
    assert len(buffer) == 1
    assert len(buffer.n_step_buffer) == 3


def test_nstep_buffer_add_with_images():
    """Test adding image transitions to MultiStepReplayBuffer."""
    buffer = MultiStepReplayBuffer(max_size=1000, n_step=3, gamma=0.99)

    # Define transition fields required for n-step returns
    buffer.reward_key = "reward"
    buffer.done_key = "done"
    buffer.ns_key = "next_state"

    # Add transitions with image observations
    for i in range(4):  # Add more than n_step
        # Create image-like tensor (C, H, W)
        img = torch.ones((3, 84, 84)) * i
        next_img = torch.ones((3, 84, 84)) * (i + 1)

        data = TensorDict(
            {
                "state": img,
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
                "next_state": next_img,
                "done": torch.tensor([False]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Should have processed at least one n-step transition
    assert len(buffer) >= 1

    # Sample and verify image dimensions
    if len(buffer) > 0:
        samples = buffer.sample(1)
        assert samples["state"].shape[1:] == (3, 84, 84)


def test_nstep_reward_calculation():
    """Test n-step reward calculation."""
    buffer = MultiStepReplayBuffer(max_size=1000, n_step=3, gamma=0.9)

    # Define required keys
    buffer.reward_key = "reward"
    buffer.done_key = "done"
    buffer.ns_key = "next_state"

    # Initialize buffer storage to avoid None reference
    buffer._storage = TensorDict({}, batch_size=[0, 0])

    # Fill n-step buffer with transitions
    for i in range(3):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i]),
                "reward": torch.tensor([float(i + 1)]),
                "next_state": torch.tensor([i + 3, i + 4, i + 5]),
                "done": torch.tensor([False]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.n_step_buffer.append(data)

    # Calculate n-step return
    n_step_data = buffer._get_n_step_info()

    # Expected reward: r1 + gamma*r2 + gamma^2*r3
    expected_reward = 1.0 + 0.9 * 2.0 + 0.9 * 0.9 * 3.0

    assert torch.isclose(n_step_data["reward"], torch.tensor(expected_reward))
    assert torch.all(n_step_data["next_state"] == buffer.n_step_buffer[2]["next_state"])
    assert torch.all(n_step_data["done"] == buffer.n_step_buffer[2]["done"])


def test_nstep_early_termination():
    """Test n-step calculation with early termination."""
    buffer = MultiStepReplayBuffer(max_size=1000, n_step=3, gamma=0.9)

    # Define required keys
    buffer.reward_key = "reward"
    buffer.done_key = "done"
    buffer.ns_key = "next_state"

    # Fill n-step buffer with transitions where second one terminates
    data1 = TensorDict(
        {
            "state": torch.tensor([1, 2, 3]),
            "action": torch.tensor([0]),
            "reward": torch.tensor([1.0]),
            "next_state": torch.tensor([4, 5, 6]),
            "done": torch.tensor([False]),
        },
    )
    data1 = data1.unsqueeze(0)
    data1.batch_size = [1]
    buffer.n_step_buffer.append(data1)

    data2 = TensorDict(
        {
            "state": torch.tensor([4, 5, 6]),
            "action": torch.tensor([1]),
            "reward": torch.tensor([2.0]),
            "next_state": torch.tensor([7, 8, 9]),
            "done": torch.tensor([True]),
        },
    )
    data2 = data2.unsqueeze(0)
    data2.batch_size = [1]
    buffer.n_step_buffer.append(data2)

    data3 = TensorDict(
        {
            "state": torch.tensor([7, 8, 9]),
            "action": torch.tensor([2]),
            "reward": torch.tensor([3.0]),
            "next_state": torch.tensor([10, 11, 12]),
            "done": torch.tensor([False]),
        },
    )
    data3 = data3.unsqueeze(0)
    data3.batch_size = [1]
    buffer.n_step_buffer.append(data3)

    # Calculate n-step return
    n_step_data = buffer._get_n_step_info()

    # Expected reward: r1 + gamma*r2 (no r3 because episode terminates)
    expected_reward = 1.0 + 0.9 * 2.0

    assert torch.isclose(n_step_data["reward"], torch.tensor(expected_reward))
    assert torch.all(n_step_data["next_state"] == data2["next_state"])
    assert torch.all(n_step_data["done"] == data2["done"])


def test_nstep_sampling_with_indices():
    """Test sampling from MultiStepReplayBuffer with indices."""
    buffer = MultiStepReplayBuffer(max_size=1000, n_step=3, gamma=0.99)

    # Define transition fields required for n-step returns
    buffer.reward_key = "reward"
    buffer.done_key = "done"
    buffer.ns_key = "next_state"

    # Add transitions to fill buffer
    for i in range(10):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
                "next_state": torch.tensor([i + 3, i + 4, i + 5]),
                "done": torch.tensor([i % 5 == 0]),  # Some episodes terminate
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Skip test if buffer is empty
    if len(buffer) == 0:
        pytest.skip("Buffer is empty, cannot test sampling")

    # Sample with indices
    samples = buffer.sample(min(3, len(buffer)), return_idx=True)

    # Verify indices are returned
    assert "idxs" in samples
    assert len(samples["idxs"]) == min(3, len(buffer))
    assert all(0 <= idx < len(buffer) for idx in samples["idxs"])


##### PrioritizedReplayBuffer class tests #####
def test_prioritized_buffer_initialization():
    """Test initialization of PrioritizedReplayBuffer."""
    max_size = 1000
    alpha = 0.6

    buffer = PrioritizedReplayBuffer(max_size, alpha)

    assert buffer.max_size == max_size
    assert buffer.alpha == alpha
    assert buffer.max_priority == 1.0
    assert buffer.tree_ptr == 0
    assert isinstance(buffer.sum_tree, SumSegmentTree)
    assert isinstance(buffer.min_tree, MinSegmentTree)


def test_prioritized_buffer_add():
    """Test adding experience to PrioritizedReplayBuffer."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add a transition
    data = TensorDict(
        {
            "state": torch.tensor([1, 2, 3]),
            "action": torch.tensor([0]),
            "reward": torch.tensor([1.0]),
        },
    )
    data = data.unsqueeze(0)
    data.batch_size = [1]
    buffer.add(data)

    assert len(buffer) == 1
    assert buffer.tree_ptr == 1
    assert buffer.sum_tree[0] == buffer.max_priority**buffer.alpha
    assert buffer.min_tree[0] == buffer.max_priority**buffer.alpha


def test_prioritized_buffer_add_batch():
    """Test adding batch of experiences to PrioritizedReplayBuffer."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add batch
    batch_data = TensorDict(
        {
            "state": torch.tensor([[1, 2, 3], [4, 5, 6]]),
            "action": torch.tensor([[0], [1]]),
            "reward": torch.tensor([[1.0], [2.0]]),
        },
    )
    batch_data.batch_size = [2]
    buffer.add(batch_data)

    assert len(buffer) == 2
    assert buffer.tree_ptr == 2
    assert buffer.sum_tree[0] == buffer.max_priority**buffer.alpha
    assert buffer.sum_tree[1] == buffer.max_priority**buffer.alpha


def test_prioritized_buffer_with_image_observations():
    """Test PrioritizedReplayBuffer with image observations."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add experiences with image observations
    for i in range(5):
        # Create image-like tensor (C, H, W)
        img = torch.ones((3, 84, 84)) * i

        data = TensorDict(
            {
                "state": img,
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Sample experiences
    batch_size = 3
    beta = 0.4
    samples = buffer.sample(batch_size, beta)

    # Verify sample structure for image observations
    assert isinstance(samples, TensorDict)
    assert "state" in samples
    assert samples["state"].shape == (batch_size, 3, 84, 84)
    assert "weights" in samples
    assert "idxs" in samples


def test_prioritized_buffer_with_dict_observations():
    """Test PrioritizedReplayBuffer with dictionary observations."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add experiences with dictionary observations
    for i in range(5):
        # Create dictionary observation
        obs = TensorDict(
            {
                "image": torch.ones((3, 84, 84)) * i,
                "vector": torch.tensor([i, i + 1, i + 2]),
            },
        )
        data = TensorDict(
            {
                "state": obs,
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Sample experiences
    batch_size = 3
    beta = 0.4
    samples = buffer.sample(batch_size, beta)

    # Verify sample structure for dictionary observations
    assert isinstance(samples, TensorDict)
    assert "state" in samples
    assert isinstance(samples["state"], TensorDict)
    assert "image" in samples["state"]
    assert "vector" in samples["state"]
    assert samples["state"]["image"].shape == (batch_size, 3, 84, 84)
    assert samples["state"]["vector"].shape == (batch_size, 3)
    assert "weights" in samples
    assert "idxs" in samples


def test_prioritized_sampling():
    """Test sampling from PrioritizedReplayBuffer with priority."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add multiple experiences
    for i in range(10):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Sample experiences
    batch_size = 5
    beta = 0.4
    samples = buffer.sample(batch_size, beta)

    # Verify sample structure
    assert isinstance(samples, TensorDict)
    assert "state" in samples
    assert "action" in samples
    assert "reward" in samples
    assert "weights" in samples
    assert "idxs" in samples
    assert samples.batch_size == torch.Size([batch_size])
    assert samples["weights"].shape == (batch_size, 1)
    assert samples["idxs"].shape == (batch_size, 1)


def test_prioritized_buffer_update_priorities():
    """Test updating priorities in PrioritizedReplayBuffer."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add experiences
    for i in range(5):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # 'Sample' experiences
    batch_size = 3
    indices = np.arange(batch_size)

    # Update priorities
    new_priorities = torch.tensor([2.0, 3.0, 4.0])
    buffer.update_priorities(indices, new_priorities)

    # Verify priorities were updated
    for idx, priority in zip(indices, new_priorities):
        assert buffer.sum_tree[idx.item()] == priority.item() ** buffer.alpha, (
            priority.item() ** buffer.alpha
        )
        assert buffer.min_tree[idx.item()] == priority.item() ** buffer.alpha, (
            priority.item() ** buffer.alpha
        )

    # Max priority should be updated
    assert buffer.max_priority == 4.0


def test_proportional_sampling():
    """Test proportional sampling based on priorities."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add experiences with different priorities
    for i in range(5):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

        # Update priorities to make them different
        buffer._update_priority(i, float(i + 1))

    # Mock random sampling to test proportional selection
    original_rand = torch.rand
    try:
        # Use deterministic values for testing
        predetermined_values = [0.1, 0.3, 0.6, 0.8]
        counter = 0

        def mock_rand(size):
            nonlocal counter
            val = predetermined_values[counter % len(predetermined_values)]
            counter += 1
            return torch.tensor([val])

        torch.rand = mock_rand

        # Sample indices
        batch_size = 4
        indices = buffer._sample_proportional(batch_size)

        # Verify indices are within range
        assert all(0 <= idx < len(buffer) for idx in indices)

    finally:
        # Restore original function
        torch.rand = original_rand


def test_weight_calculation():
    """Test weight calculation for importance sampling."""
    buffer = PrioritizedReplayBuffer(max_size=100, alpha=0.6)

    # Add experiences
    for i in range(5):
        data = TensorDict(
            {
                "state": torch.tensor([i, i + 1, i + 2]),
                "action": torch.tensor([i % 3]),
                "reward": torch.tensor([float(i)]),
            },
        )
        data = data.unsqueeze(0)
        data.batch_size = [1]
        buffer.add(data)

    # Update priorities to known values
    priorities = [0.5, 1.0, 1.5, 2.0, 2.5]
    for i, priority in enumerate(priorities):
        buffer._update_priority(i, priority)

    # Test weight calculation with beta=0.4
    beta = 0.4
    for i, priority in enumerate(priorities):
        weight = buffer._calculate_weights(torch.tensor([i]), beta)[0]

        # Calculate expected weight:
        # w = (p * N)^(-beta) / max_weight where max_weight = (p_min * N)^(-beta)
        p_sample = (priority**buffer.alpha) / buffer.sum_tree.sum()
        p_min = buffer.min_tree.min() / buffer.sum_tree.sum()
        expected_weight = ((p_sample * len(buffer)) ** -beta) / (
            (p_min * len(buffer)) ** -beta
        )

        assert torch.isclose(weight, torch.tensor(expected_weight), rtol=1e-5)



import copy

import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces

from agilerl.algorithms.cqn import CQN
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import assert_not_equal_state_dict, assert_state_dicts_equal


class DummyCQN(CQN):
    def __init__(self, observation_space, action_space, *args, **kwargs):
        super().__init__(observation_space, action_space, *args, **kwargs)

        self.tensor_test = torch.randn(1)


class DummyEnv:
    def __init__(self, state_size, vect=True, num_envs=2):
        self.state_size = state_size
        self.vect = vect
        if self.vect:
            self.state_size = (num_envs,) + self.state_size
            self.n_envs = num_envs
            self.num_envs = num_envs
        else:
            self.n_envs = 1

    def reset(self):
        return np.random.rand(*self.state_size), {}

    def step(self, action):
        return (
            np.random.rand(*self.state_size),
            np.random.randint(0, 5, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            {},
        )


@pytest.fixture(scope="module")
def simple_mlp():
    network = nn.Sequential(
        nn.Linear(4, 20),
        nn.ReLU(),
        nn.Linear(20, 10),
        nn.ReLU(),
        nn.Linear(10, 1),
        nn.Tanh(),
    )
    return network


@pytest.fixture(scope="module")
def simple_cnn():
    network = nn.Sequential(
        nn.Conv2d(
            3, 16, kernel_size=3, stride=1, padding=1
        ),  # Input channels: 3 (for RGB images), Output channels: 16
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Conv2d(
            16, 32, kernel_size=3, stride=1, padding=1
        ),  # Input channels: 16, Output channels: 32
        nn.ReLU(),
        nn.MaxPool2d(kernel_size=2, stride=2),
        nn.Flatten(),  # Flatten the 2D feature map to a 1D vector
        nn.Linear(32 * 16 * 16, 128),  # Fully connected layer with 128 output features
        nn.ReLU(),
        nn.Linear(128, 1),  # Output layer with num_classes output features
    )
    return network


# initialize CQN with valid parameters
@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
        ("image_space", EvolvableCNN),
        ("dict_space", EvolvableMultiInput),
        ("multidiscrete_space", EvolvableMLP),
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_cqn(observation_space, encoder_cls, accelerator, request):
    action_space = spaces.Discrete(2)
    observation_space = request.getfixturevalue(observation_space)
    cqn = CQN(observation_space, action_space, accelerator=accelerator)

    expected_device = accelerator.device if accelerator else "cpu"
    assert cqn.observation_space == observation_space
    assert cqn.action_space == action_space
    assert cqn.batch_size == 64
    assert cqn.lr == 0.0001
    assert cqn.learn_step == 5
    assert cqn.gamma == 0.99
    assert cqn.tau == 0.001
    assert cqn.mut is None
    assert cqn.device == expected_device
    assert cqn.accelerator == accelerator
    assert cqn.index == 0
    assert cqn.scores == []
    assert cqn.fitness == []
    assert cqn.steps == [0]
    assert isinstance(cqn.actor.encoder, encoder_cls)
    assert isinstance(cqn.actor_target.encoder, encoder_cls)
    expected_opt_cls = AcceleratedOptimizer if accelerator else optim.Adam
    assert isinstance(cqn.optimizer.optimizer, expected_opt_cls)
    assert isinstance(cqn.criterion, nn.MSELoss)
    cqn.clean_up()


# Can initialize cqn with an actor network
# TODO: This will be deprecated in the future
@pytest.mark.parametrize(
    "observation_space, actor_network, input_tensor",
    [
        ("vector_space", "simple_mlp", torch.randn(1, 4)),
        (
            "image_space",
            "simple_cnn",
            torch.randn(1, 3, 64, 64),
        ),
    ],
)
def test_initialize_cqn_with_make_evo(
    observation_space, actor_network, input_tensor, request
):
    action_space = spaces.Discrete(2)
    observation_space = request.getfixturevalue(observation_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)

    cqn = CQN(observation_space, action_space, actor_network=actor_network)

    assert cqn.observation_space == observation_space
    assert cqn.action_space == action_space
    assert cqn.batch_size == 64
    assert cqn.lr == 0.0001
    assert cqn.learn_step == 5
    assert cqn.gamma == 0.99
    assert cqn.tau == 0.001
    assert cqn.mut is None
    assert cqn.device == "cpu"
    assert cqn.accelerator is None
    assert cqn.index == 0
    assert cqn.scores == []
    assert cqn.fitness == []
    assert cqn.steps == [0]
    assert isinstance(cqn.optimizer.optimizer, optim.Adam)
    assert isinstance(cqn.criterion, nn.MSELoss)
    cqn.clean_up()


@pytest.mark.parametrize(
    "observation_space, net_type",
    [
        ("vector_space", "mlp"),
        ("image_space", "cnn"),
    ],
)
def test_initialize_cqn_with_actor_network_evo_net(
    observation_space, net_type, request
):
    action_space = spaces.Discrete(2)
    observation_space = request.getfixturevalue(observation_space)
    if net_type == "mlp":
        actor_network = EvolvableMLP(
            num_inputs=observation_space.shape[0],
            num_outputs=action_space.n,
            hidden_size=[64, 64],
            activation="ReLU",
        )
    else:
        actor_network = EvolvableCNN(
            input_shape=observation_space.shape,
            num_outputs=action_space.n,
            channel_size=[8, 8],
            kernel_size=[2, 2],
            stride_size=[1, 1],
            activation="ReLU",
        )

    cqn = CQN(observation_space, action_space, actor_network=actor_network)

    assert cqn.observation_space == observation_space
    assert cqn.action_space == action_space
    assert cqn.batch_size == 64
    assert cqn.lr == 0.0001
    assert cqn.learn_step == 5
    assert cqn.gamma == 0.99
    assert cqn.tau == 0.001
    assert cqn.mut is None
    assert cqn.device == "cpu"
    assert cqn.accelerator is None
    assert cqn.index == 0
    assert cqn.scores == []
    assert cqn.fitness == []
    assert cqn.steps == [0]
    assert isinstance(cqn.optimizer.optimizer, optim.Adam)
    assert isinstance(cqn.criterion, nn.MSELoss)
    cqn.clean_up()


def test_init_with_incorrect_actor_net(vector_space):
    action_space = spaces.Discrete(2)
    actor_network = "String"

    with pytest.raises(TypeError) as e:
        cqn = CQN(vector_space, action_space, actor_network=actor_network)
        assert cqn
        assert (
            e
            == f"'actor_network' argument is of type {type(actor_network)}, but must be of type nn.Module."
        )


# Returns the expected action when given a state observation and epsilon=0 or 1.
def test_returns_expected_action_epsilon_greedy(vector_space):
    action_space = spaces.Discrete(2)

    cqn = CQN(vector_space, action_space)
    state = np.array([1, 2, 3, 4])

    action_mask = None

    epsilon = 0
    action = cqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action >= 0 and action < action_space.n

    epsilon = 1
    action = cqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action >= 0 and action < action_space.n
    cqn.clean_up()


# Returns the expected action when given a state observation and action mask.
def test_returns_expected_action_mask(vector_space):
    accelerator = Accelerator()
    action_space = spaces.Discrete(2)

    cqn = CQN(vector_space, action_space, accelerator=accelerator)
    state = np.array([1, 2, 3, 4])

    action_mask = np.array([0, 1])

    epsilon = 0
    action = cqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action == 1

    epsilon = 1
    action = cqn.get_action(state, epsilon, action_mask)[0]

    assert action.is_integer()
    assert action == 1
    cqn.clean_up()


# learns from experiences and updates network parameters
def test_learns_from_experiences(vector_space):
    action_space = spaces.Discrete(2)
    batch_size = 64

    # Create an instance of the cqn class
    cqn = CQN(vector_space, action_space, batch_size=batch_size)

    # Create a batch of experiences
    states = torch.randn(batch_size, vector_space.shape[0])
    actions = torch.randint(0, action_space.n, (batch_size, 1))
    rewards = torch.randn((batch_size, 1))
    next_states = torch.randn(batch_size, vector_space.shape[0])
    dones = torch.randint(0, 2, (batch_size, 1))

    experiences = [states, actions, rewards, next_states, dones]

    # Copy state dict before learning - should be different to after updating weights
    actor = cqn.actor
    actor_target = cqn.actor_target
    actor_pre_learn_sd = copy.deepcopy(cqn.actor.state_dict())
    actor_target_pre_learn_sd = copy.deepcopy(cqn.actor_target.state_dict())

    # Call the learn method
    cqn.learn(experiences)

    assert actor == cqn.actor
    assert actor_target == cqn.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, cqn.actor.state_dict())
    assert_not_equal_state_dict(
        actor_target_pre_learn_sd, cqn.actor_target.state_dict()
    )
    cqn.clean_up()


# handles double Q-learning
def test_handles_double_q_learning(discrete_space):
    accelerator = Accelerator()
    action_space = spaces.Discrete(2)
    double = True
    batch_size = 64

    # Create an instance of the cqn class
    cqn = CQN(
        discrete_space,
        action_space,
        double=double,
        batch_size=batch_size,
        accelerator=accelerator,
    )

    # Create a batch of experiences
    states = torch.randint(0, discrete_space.n, (batch_size, 1))
    actions = torch.randint(0, action_space.n, (batch_size, 1))
    rewards = torch.randn((batch_size, 1))
    next_states = torch.randint(0, discrete_space.n, (batch_size, 1))
    dones = torch.randint(0, 2, (batch_size, 1))

    experiences = [states, actions, rewards, next_states, dones]

    # Copy state dict before learning - should be different to after updating weights
    actor = cqn.actor
    actor_target = cqn.actor_target
    actor_pre_learn_sd = copy.deepcopy(cqn.actor.state_dict())
    actor_target_pre_learn_sd = copy.deepcopy(cqn.actor_target.state_dict())

    # Call the learn method
    cqn.learn(experiences)

    assert actor == cqn.actor
    assert actor_target == cqn.actor_target
    assert_not_equal_state_dict(actor_pre_learn_sd, cqn.actor.state_dict())
    assert_not_equal_state_dict(
        actor_target_pre_learn_sd, cqn.actor_target.state_dict()
    )
    cqn.clean_up()


# Updates target network parameters with soft update
def test_soft_update(vector_space):
    action_space = spaces.Discrete(2)
    net_config = {"encoder_config": {"hidden_size": [64, 64]}}
    batch_size = 64
    lr = 1e-4
    learn_step = 5
    gamma = 0.99
    tau = 1e-3
    mut = None
    double = False
    actor_network = None
    device = "cpu"
    accelerator = None
    wrap = True

    cqn = CQN(
        vector_space,
        action_space,
        net_config=net_config,
        batch_size=batch_size,
        lr=lr,
        learn_step=learn_step,
        gamma=gamma,
        tau=tau,
        mut=mut,
        double=double,
        actor_network=actor_network,
        device=device,
        accelerator=accelerator,
        wrap=wrap,
    )

    cqn.soft_update()

    eval_params = list(cqn.actor.parameters())
    target_params = list(cqn.actor_target.parameters())
    expected_params = [
        cqn.tau * eval_param + (1.0 - cqn.tau) * target_param
        for eval_param, target_param in zip(eval_params, target_params)
    ]

    assert all(
        torch.allclose(expected_param, target_param)
        for expected_param, target_param in zip(expected_params, target_params)
    )
    cqn.clean_up()


# Runs algorithm test loop
@pytest.mark.parametrize("observation_space", ["vector_space", "image_space"])
@pytest.mark.parametrize("num_envs", [1, 3])
def test_algorithm_test_loop(observation_space, num_envs, request):
    observation_space = request.getfixturevalue(observation_space)
    action_space = spaces.Discrete(2)

    vect = num_envs > 1
    env = DummyEnv(state_size=observation_space.shape, vect=vect, num_envs=num_envs)
    agent = CQN(observation_space=observation_space, action_space=action_space)
    mean_score = agent.test(env, max_steps=10)
    assert isinstance(mean_score, float)
    agent.clean_up()


# Clones the agent and returns an identical agent.
@pytest.mark.parametrize("observation_space", ["vector_space"])
def test_clone_returns_identical_agent(observation_space, request):
    action_space = spaces.Discrete(2)
    observation_space = request.getfixturevalue(observation_space)

    cqn = DummyCQN(observation_space, action_space)
    cqn.tensor_attribute = torch.randn(1)
    clone_agent = cqn.clone()

    assert clone_agent.observation_space == cqn.observation_space
    assert clone_agent.action_space == cqn.action_space
    assert clone_agent.batch_size == cqn.batch_size
    assert clone_agent.lr == cqn.lr
    assert clone_agent.learn_step == cqn.learn_step
    assert clone_agent.gamma == cqn.gamma
    assert clone_agent.tau == cqn.tau
    assert clone_agent.mut == cqn.mut
    assert clone_agent.device == cqn.device
    assert clone_agent.accelerator == cqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), cqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), cqn.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), cqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == cqn.fitness
    assert clone_agent.steps == cqn.steps
    assert clone_agent.scores == cqn.scores
    assert clone_agent.tensor_attribute == cqn.tensor_attribute
    assert clone_agent.tensor_test == cqn.tensor_test
    cqn.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    cqn = CQN(observation_space, action_space, accelerator=accelerator)
    clone_agent = cqn.clone()

    assert clone_agent.observation_space == cqn.observation_space
    assert clone_agent.action_space == cqn.action_space
    # assert clone_agent.actor_network == cqn.actor_network
    assert clone_agent.batch_size == cqn.batch_size
    assert clone_agent.lr == cqn.lr
    assert clone_agent.learn_step == cqn.learn_step
    assert clone_agent.gamma == cqn.gamma
    assert clone_agent.tau == cqn.tau
    assert clone_agent.mut == cqn.mut
    assert clone_agent.device == cqn.device
    assert clone_agent.accelerator == cqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), cqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), cqn.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), cqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == cqn.fitness
    assert clone_agent.steps == cqn.steps
    assert clone_agent.scores == cqn.scores
    cqn.clean_up()
    clone_agent.clean_up()

    accelerator = Accelerator()
    cqn = CQN(observation_space, action_space, accelerator=accelerator, wrap=False)
    clone_agent = cqn.clone(wrap=False)

    assert clone_agent.observation_space == cqn.observation_space
    assert clone_agent.action_space == cqn.action_space
    # assert clone_agent.actor_network == cqn.actor_network
    assert clone_agent.batch_size == cqn.batch_size
    assert clone_agent.lr == cqn.lr
    assert clone_agent.learn_step == cqn.learn_step
    assert clone_agent.gamma == cqn.gamma
    assert clone_agent.tau == cqn.tau
    assert clone_agent.mut == cqn.mut
    assert clone_agent.device == cqn.device
    assert clone_agent.accelerator == cqn.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), cqn.actor.state_dict())
    assert_state_dicts_equal(
        clone_agent.actor_target.state_dict(), cqn.actor_target.state_dict()
    )
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), cqn.optimizer.state_dict()
    )
    assert clone_agent.fitness == cqn.fitness
    assert clone_agent.steps == cqn.steps
    assert clone_agent.scores == cqn.scores
    cqn.clean_up()
    clone_agent.clean_up()


def test_clone_new_index(vector_space):
    action_space = spaces.Discrete(2)

    cqn = CQN(vector_space, action_space)
    clone_agent = cqn.clone(index=100)

    assert clone_agent.index == 100
    cqn.clean_up()
    clone_agent.clean_up()



from dataclasses import asdict

import pytest
import torch
import torch.nn.functional as F
from gymnasium import spaces

from agilerl.modules import (
    EvolvableCNN,
    EvolvableLSTM,
    EvolvableMLP,
    EvolvableModule,
    EvolvableMultiInput,
    EvolvableSimBa,
)
from agilerl.modules.configs import MlpNetConfig
from agilerl.networks.base import EvolvableNetwork
from agilerl.networks.value_networks import ValueNetwork
from tests.helper_functions import (
    assert_close_dict,
    assert_not_equal_state_dict,
    assert_state_dicts_equal,
    check_equal_params_ind,
)


@pytest.fixture(scope="module")
def head_config():
    return asdict(MlpNetConfig(hidden_size=[64, 64]))


@pytest.mark.parametrize(
    "observation_space, encoder_type",
    [
        ("dict_space", "multi_input"),
        ("discrete_space", "mlp"),
        ("vector_space", "mlp"),
        ("image_space", "cnn"),
    ],
)
def test_value_function_initialization(observation_space, encoder_type, request):
    observation_space = request.getfixturevalue(observation_space)

    network = ValueNetwork(observation_space)

    assert network.observation_space == observation_space

    if encoder_type == "multi_input":
        assert isinstance(network.encoder, EvolvableMultiInput)
    elif encoder_type == "mlp":
        assert isinstance(network.encoder, EvolvableMLP)
    elif encoder_type == "cnn":
        assert isinstance(network.encoder, EvolvableCNN)

    evolvable_modules = network.modules()
    assert "encoder" in evolvable_modules
    assert "head_net" in evolvable_modules


def test_value_function_initialization_recurrent(vector_space):
    observation_space = vector_space
    network = ValueNetwork(observation_space, recurrent=True)

    assert network.observation_space == observation_space
    assert isinstance(network.encoder, EvolvableLSTM)

    evolvable_modules = network.modules()
    assert "encoder" in evolvable_modules
    assert "head_net" in evolvable_modules


def test_value_function_initialization_simba(vector_space):
    network = ValueNetwork(vector_space, simba=True)

    assert network.observation_space == vector_space
    assert isinstance(network.encoder, EvolvableSimBa)

    evolvable_modules = network.modules()
    assert "encoder" in evolvable_modules
    assert "head_net" in evolvable_modules


@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
def test_value_function_mutation_methods(
    observation_space, head_config, dummy_rng, request
):
    observation_space = request.getfixturevalue(observation_space)
    network = ValueNetwork(observation_space, head_config=head_config)
    network.rng = dummy_rng

    for method in network.mutation_methods:
        new_network = network.clone()
        getattr(new_network, method)()

        if "." in method:
            net_name = method.split(".")[0]
            mutated_module: EvolvableModule = getattr(new_network, net_name)
            exec_method = new_network.last_mutation_attr.split(".")[-1]

            if isinstance(observation_space, (spaces.Tuple, spaces.Dict)):
                mutated_attr = mutated_module.last_mutation_attr.split(".")[-1]
            else:
                mutated_attr = mutated_module.last_mutation_attr

            assert mutated_attr == exec_method

        if new_network.last_mutation_attr is not None:
            # Check that architecture has changed
            assert_not_equal_state_dict(network.state_dict(), new_network.state_dict())

            # Checks that parameters that are not mutated are the same
            check_equal_params_ind(network, new_network)
        else:
            raise ValueError(
                f"Last mutation attribute is None. Expected {method} to be applied."
            )


@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
def test_value_function_forward(observation_space: spaces.Space, request):
    observation_space = request.getfixturevalue(observation_space)

    network = ValueNetwork(observation_space)

    x_np = observation_space.sample()

    if isinstance(observation_space, spaces.Discrete):
        x_np = (
            F.one_hot(torch.tensor(x_np), num_classes=observation_space.n)
            .float()
            .numpy()
        )

    with torch.no_grad():
        out = network(x_np)

    assert isinstance(out, torch.Tensor)
    assert out.shape == torch.Size([1, 1])

    if isinstance(observation_space, spaces.Dict):
        x = {key: torch.tensor(value) for key, value in x_np.items()}
    elif isinstance(observation_space, spaces.Tuple):
        x = tuple(torch.tensor(value) for value in x_np)
    else:
        x = torch.tensor(x_np)

    with torch.no_grad():
        out = network(x)

    assert isinstance(out, torch.Tensor)
    assert out.shape == torch.Size([1, 1])


@pytest.mark.parametrize(
    "observation_space", ["dict_space", "discrete_space", "vector_space", "image_space"]
)
def test_value_function_clone(observation_space: spaces.Space, request):
    observation_space = request.getfixturevalue(observation_space)

    network = ValueNetwork(observation_space)

    original_net_dict = dict(network.named_parameters())
    clone = network.clone()
    assert isinstance(clone, EvolvableNetwork)

    assert_close_dict(network.init_dict, clone.init_dict)

    assert_state_dicts_equal(clone.state_dict(), network.state_dict())
    for key, param in clone.named_parameters():
        torch.testing.assert_close(param, original_net_dict[key])



import copy

import gymnasium
import numpy as np
import pytest
import torch
import torch.nn as nn
import torch.optim as optim
from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from gymnasium import spaces

from agilerl.algorithms.ppo import PPO
from agilerl.components.rollout_buffer import RolloutBuffer
from agilerl.modules import EvolvableCNN, EvolvableMLP, EvolvableMultiInput
from agilerl.rollouts import collect_rollouts, collect_rollouts_recurrent
from agilerl.typing import BPTTSequenceType
from agilerl.wrappers.make_evolvable import MakeEvolvable
from tests.helper_functions import assert_not_equal_state_dict, assert_state_dicts_equal


def get_batch_states(observation_space, num_steps) -> tuple[torch.Tensor, torch.Tensor]:
    if isinstance(observation_space, spaces.Discrete):
        states = torch.randint(0, observation_space.n, (num_steps,)).float()
        next_states = torch.randint(0, observation_space.n, (1,)).float()
    elif isinstance(observation_space, spaces.MultiDiscrete):
        states = torch.stack(
            [
                torch.randint(0, high, (num_steps,))
                for high in observation_space.nvec.tolist()
            ],
            dim=1,
        ).float()
        next_states = torch.stack(
            [torch.randint(0, high, (1,)) for high in observation_space.nvec.tolist()],
            dim=1,
        ).float()
    elif isinstance(observation_space, spaces.MultiBinary):
        states = torch.randint(0, 2, (num_steps,)).float()
        next_states = torch.randint(0, 2, (1,)).float()
    elif isinstance(observation_space, spaces.Box):
        states = torch.rand(num_steps, *observation_space.shape)
        next_states = torch.rand(1, *observation_space.shape)
    elif isinstance(observation_space, spaces.Dict):
        states = {
            key: torch.rand(num_steps, *space.shape)
            for key, space in observation_space.spaces.items()
        }
        next_states = {
            key: torch.rand(1, *space.shape)
            for key, space in observation_space.spaces.items()
        }
    elif isinstance(observation_space, spaces.Tuple):
        states = tuple(
            torch.rand(num_steps, *space.shape) for space in observation_space.spaces
        )
        next_states = tuple(
            torch.rand(1, *space.shape) for space in observation_space.spaces
        )
    return states, next_states


class DummyPPO(PPO):
    def __init__(self, observation_space, action_space, *args, **kwargs):
        super().__init__(observation_space, action_space, *args, **kwargs)

        self.tensor_test = torch.randn(1)


class DummyEnv:
    def __init__(self, state_size, vect=True, num_envs=2):
        self.state_size = state_size
        self.vect = vect
        self.num_envs = num_envs
        if self.vect:
            self.state_size = (num_envs,) + self.state_size
            self.n_envs = num_envs
        else:
            self.n_envs = 1

    def reset(self):
        return np.random.rand(*self.state_size), {}

    def step(self, action):
        return (
            np.random.rand(*self.state_size),
            np.random.randint(0, 5, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            np.random.randint(0, 2, self.n_envs),
            {},
        )


class SimpleCNN(nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(
            3, 16, kernel_size=3, stride=1, padding=1
        )  # Input channels: 3 (for RGB images), Output channels: 16
        self.relu1 = nn.ReLU()
        self.mp1 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.conv2 = nn.Conv2d(
            16, 32, kernel_size=3, stride=1, padding=1
        )  # Input channels: 16, Output channels: 32
        self.relu2 = nn.ReLU()
        self.mp2 = nn.MaxPool2d(kernel_size=2, stride=2)
        self.flat = nn.Flatten()  # Flatten the 2D feature map to a 1D vector
        self.linear1 = nn.Linear(
            32 * 16 * 16, 128
        )  # Fully connected layer with 128 output features
        self.relu3 = nn.ReLU()
        self.linear2 = nn.Linear(
            128, 128
        )  # Fully connected layer with 128 output features

    def forward(self, x, xc):
        x = self.mp1(self.relu1(self.conv1(x)))
        x = self.mp2(self.relu2(self.conv2(x)))
        x = self.flat(x)
        x = self.relu3(self.linear1(x))
        x = self.relu3(self.linear2(x))
        return x


@pytest.fixture(scope="function")
def build_ppo(observation_space, action_space, recurrent, accelerator, request):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)
    use_rollout_buffer = recurrent
    return PPO(
        observation_space,
        action_space,
        use_rollout_buffer=use_rollout_buffer,
        recurrent=recurrent,
        accelerator=accelerator,
    )


# Initializes all necessary attributes with default values
@pytest.mark.parametrize(
    "observation_space, encoder_cls",
    [
        ("vector_space", EvolvableMLP),
        ("image_space", EvolvableCNN),
        ("dict_space", EvolvableMultiInput),
    ],
)
@pytest.mark.parametrize(
    "action_space",
    [
        "vector_space",
        "discrete_space",
        "multidiscrete_space",
        "multibinary_space",
    ],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_initialize_ppo(
    observation_space, action_space, encoder_cls, accelerator, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)
    ppo = PPO(
        observation_space,
        action_space,
        accelerator=accelerator,
        use_rollout_buffer=False,
        recurrent=False,
    )

    assert ppo.algo == "PPO"
    assert ppo.observation_space == observation_space
    assert ppo.action_space == action_space
    assert ppo.batch_size == 64
    assert ppo.lr == 1e-4
    assert ppo.gamma == 0.99
    assert ppo.gae_lambda == 0.95
    assert ppo.mut is None
    assert ppo.action_std_init == 0.0
    assert ppo.clip_coef == 0.2
    assert ppo.ent_coef == 0.01
    assert ppo.vf_coef == 0.5
    assert ppo.max_grad_norm == 0.5
    assert ppo.target_kl is None
    assert ppo.update_epochs == 4
    assert ppo.device == accelerator.device if accelerator else "cpu"
    assert ppo.accelerator == accelerator
    assert ppo.index == 0
    assert ppo.scores == []
    assert ppo.fitness == []
    assert ppo.steps == [0]
    assert isinstance(ppo.actor.encoder, encoder_cls)
    assert isinstance(ppo.critic.encoder, encoder_cls)
    expected_optimizer = AcceleratedOptimizer if accelerator else optim.Adam
    assert isinstance(ppo.optimizer.optimizer, expected_optimizer)
    ppo.clean_up()


# Can initialize ppo with an actor network
# TODO: Will be deprecated in the future
@pytest.mark.parametrize(
    "obs_space, action_space, actor_network, critic_network, input_tensor, input_tensor_critic",
    [
        (
            "vector_space",
            "discrete_space",
            "simple_mlp",
            "simple_mlp_critic",
            torch.randn(1, 4),
            torch.randn(1, 6),
        ),
    ],
)
def test_initialize_ppo_with_make_evo(
    obs_space,
    action_space,
    actor_network,
    critic_network,
    input_tensor,
    input_tensor_critic,
    request,
):
    obs_space = request.getfixturevalue(obs_space)
    action_space = request.getfixturevalue(action_space)
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)
    critic_network = request.getfixturevalue(critic_network)
    critic_network = MakeEvolvable(critic_network, input_tensor_critic)

    ppo = PPO(
        obs_space,
        action_space,
        actor_network=actor_network,
        critic_network=critic_network,
    )

    assert ppo.observation_space == obs_space
    assert ppo.action_space == action_space
    assert ppo.batch_size == 64
    assert ppo.lr == 1e-4
    assert ppo.gamma == 0.99
    assert ppo.gae_lambda == 0.95
    assert ppo.mut is None
    assert ppo.action_std_init == 0.0
    assert ppo.clip_coef == 0.2
    assert ppo.ent_coef == 0.01
    assert ppo.vf_coef == 0.5
    assert ppo.max_grad_norm == 0.5
    assert ppo.target_kl is None
    assert ppo.update_epochs == 4
    assert ppo.device == "cpu"
    assert ppo.accelerator is None
    assert ppo.index == 0
    assert ppo.scores == []
    assert ppo.fitness == []
    assert ppo.steps == [0]
    assert isinstance(ppo.optimizer.optimizer, optim.Adam)
    assert ppo.num_envs == 1
    ppo.clean_up()


def test_initialize_ppo_with_incorrect_actor_net(vector_space, discrete_space):
    actor_network = "dummy"
    critic_network = "dummy"
    with pytest.raises(TypeError):
        ppo = PPO(
            vector_space,
            discrete_space,
            actor_network=actor_network,
            critic_network=critic_network,
        )
        assert ppo


# Can initialize ppo with an actor network but no critic - should trigger warning
@pytest.mark.parametrize(
    "observation_space, actor_network, critic_network, input_tensor, input_tensor_critic",
    [
        (
            "vector_space",
            "simple_mlp",
            "simple_mlp_critic",
            torch.randn(1, 4),
            torch.randn(1, 6),
        ),
    ],
)
def test_initialize_ppo_with_actor_network_no_critic(
    observation_space,
    discrete_space,
    actor_network,
    critic_network,
    input_tensor,
    input_tensor_critic,
    request,
):
    actor_network = request.getfixturevalue(actor_network)
    actor_network = MakeEvolvable(actor_network, input_tensor)
    observation_space = request.getfixturevalue(observation_space)
    with pytest.raises(TypeError):
        ppo = PPO(
            observation_space,
            discrete_space,
            actor_network=actor_network,
            critic_network=critic_network,
        )
        assert ppo


@pytest.mark.parametrize(
    "observation_space", ["vector_space", "image_space", "dict_space"]
)
@pytest.mark.parametrize(
    "action_space",
    ["vector_space", "discrete_space", "multidiscrete_space", "multibinary_space"],
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
@pytest.mark.parametrize("recurrent", [False])
# Returns the expected action when given a state observation.
def test_returns_expected_action(
    observation_space, action_space, build_ppo, request, recurrent
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)
    state = observation_space.sample()

    act_ret = build_ppo.get_action(state)
    action, action_logprob, dist_entropy, state_values = act_ret

    assert isinstance(action, np.ndarray)
    assert isinstance(action_logprob, np.ndarray)
    assert isinstance(dist_entropy, np.ndarray)
    assert isinstance(state_values, np.ndarray)

    if isinstance(action_space, spaces.Discrete):
        for act in action:
            assert act.is_integer()
            assert act >= 0 and act < action_space.n
    elif isinstance(action_space, spaces.MultiDiscrete):
        assert len(action[0]) == len(action_space.nvec)
        for i, act in enumerate(action[0]):
            assert act.is_integer()
            assert act >= 0 and act < action_space.nvec[i]
    elif isinstance(action_space, spaces.MultiBinary):
        assert len(action[0]) == action_space.n
        for act in action[0]:
            assert isinstance(act, np.float32)
    else:
        assert isinstance(action, np.ndarray)
        assert action.shape == (1, *action_space.shape)

    # Now with grad=True, and eval_action
    eval_action = torch.Tensor([[0, 1, 0, 1]]).to(build_ppo.device)
    action_logprob, dist_entropy, state_values = build_ppo.evaluate_actions(
        obs=state, actions=eval_action
    )

    assert isinstance(action_logprob, torch.Tensor)
    assert isinstance(dist_entropy, torch.Tensor)
    assert isinstance(state_values, torch.Tensor)
    build_ppo.clean_up()


@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "multidiscrete_space",
        "multibinary_space",
    ],
)
@pytest.mark.parametrize(
    "action_space",
    [
        "vector_space",
        "discrete_space",
        "multidiscrete_space",
        "multibinary_space",
    ],
)
@pytest.mark.parametrize("recurrent", [True])
@pytest.mark.parametrize("accelerator", [None])
# Returns the expected action when given a state observation.
def test_returns_expected_action_recurrent(
    observation_space, action_space, recurrent, accelerator, build_ppo, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)
    state = observation_space.sample()

    act_ret = build_ppo.get_action(
        state, hidden_state=build_ppo.get_initial_hidden_state()
    )
    action, action_logprob, dist_entropy, state_values, hidden_state = act_ret

    assert isinstance(action, np.ndarray)
    assert isinstance(action_logprob, np.ndarray)
    assert isinstance(dist_entropy, np.ndarray)
    assert isinstance(state_values, np.ndarray)
    assert hidden_state is not None

    if isinstance(action_space, spaces.Discrete):
        for act in action:
            assert act.is_integer()
            assert act >= 0 and act < action_space.n
    elif isinstance(action_space, spaces.MultiDiscrete):
        assert len(action[0]) == len(action_space.nvec)
        for i, act in enumerate(action[0]):
            assert act.is_integer()
            assert act >= 0 and act < action_space.nvec[i]
    elif isinstance(action_space, spaces.MultiBinary):
        assert len(action[0]) == action_space.n
        for act in action[0]:
            assert isinstance(act, np.float32)
    else:
        assert isinstance(action, np.ndarray)
        assert action.shape == (1, *action_space.shape)

    eval_action = torch.Tensor([[0, 1, 0, 1]]).to(build_ppo.device)
    action_logprob, dist_entropy, state_values = build_ppo.evaluate_actions(
        obs=state,
        actions=eval_action,
        hidden_state=hidden_state,
    )

    assert isinstance(action_logprob, torch.Tensor)
    assert isinstance(dist_entropy, torch.Tensor)
    assert isinstance(state_values, torch.Tensor)
    build_ppo.clean_up()


def test_ppo_optimizer_parameters(vector_space, discrete_space):
    ppo = PPO(vector_space, discrete_space)

    # Store initial parameters
    initial_params = {
        name: param.clone() for name, param in ppo.actor.named_parameters()
    }

    # Perform a dummy optimization step
    dummy_input = torch.randn(1, 4)
    dummy_action = torch.tensor([0])
    dummy_log_prob = torch.tensor([1.0])

    _, _, _ = ppo.actor(dummy_input)
    loss = (dummy_log_prob - ppo.actor.action_log_prob(dummy_action)) ** 2
    loss = loss.mean()
    ppo.optimizer.zero_grad()
    loss.backward()
    ppo.optimizer.step()

    # Check if parameters have changed
    not_updated = []
    for name, param in ppo.actor.named_parameters():
        if torch.equal(initial_params[name], param):
            not_updated.append(name)

    assert not not_updated, f"The following parameters weren't updated:\n{not_updated}"
    ppo.clean_up()


@pytest.mark.parametrize("observation_space", ["vector_space"])
@pytest.mark.parametrize("action_space", ["discrete_space"])
@pytest.mark.parametrize("accelerator", [None])
@pytest.mark.parametrize("recurrent", [False])
def test_returns_expected_action_mask_vectorized(
    build_ppo, observation_space, action_space, recurrent, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)
    state = np.stack([observation_space.sample(), observation_space.sample()])
    action_mask = np.stack([np.array([0, 1]), np.array([1, 0])])
    action, _, _, _ = build_ppo.get_action(state, action_mask=action_mask)
    assert np.array_equal(action, [1, 0]), action
    build_ppo.clean_up()


@pytest.mark.parametrize(
    "observation_space", ["vector_space", "image_space", "dict_space"]
)
@pytest.mark.parametrize("accelerator", [None, Accelerator()])
def test_learns_from_experiences(
    observation_space, discrete_space, accelerator, request
):
    batch_size = 45
    observation_space = request.getfixturevalue(observation_space)
    ppo = PPO(
        observation_space=observation_space,
        action_space=discrete_space,
        batch_size=batch_size,
        accelerator=accelerator,
    )

    # Copy state dict before learning - should be different to after updating weights
    actor = ppo.actor
    actor_pre_learn_sd = copy.deepcopy(ppo.actor.state_dict())

    # Create batch size + 1 samples to ensure we can handle this
    num_steps = batch_size + 1

    # Create a batch of experiences
    states, next_states = get_batch_states(observation_space, num_steps)

    # Create a batch of experiences
    actions = torch.randint(0, discrete_space.n, (num_steps,)).float()
    log_probs = torch.randn(num_steps)
    rewards = torch.randn(num_steps)
    dones = torch.randint(0, 2, (num_steps,))
    values = torch.randn(num_steps)
    next_done = torch.zeros(1)
    experiences = [
        [states],
        [actions],
        [log_probs],
        [rewards],
        [dones],
        [values],
        [next_states],
        [next_done],
    ]

    # Call the learn method
    loss = ppo.learn(experiences)

    assert isinstance(loss, float)
    assert loss >= 0.0
    assert actor == ppo.actor
    assert_not_equal_state_dict(actor_pre_learn_sd, ppo.actor.state_dict())
    ppo.clean_up()


# Runs algorithm test loop
@pytest.mark.parametrize(
    "observation_space",
    ["vector_space", "discrete_space", "image_space"],
)
@pytest.mark.parametrize("num_envs", [1, 3])
def test_algorithm_test_loop(observation_space, discrete_space, num_envs, request):
    observation_space = request.getfixturevalue(observation_space)

    # Create a vectorised environment & test loop
    vect = num_envs > 1
    env = DummyEnv(state_size=observation_space.shape, vect=vect, num_envs=num_envs)
    agent = PPO(observation_space=observation_space, action_space=discrete_space)
    mean_score = agent.test(env, max_steps=10)
    assert isinstance(mean_score, float)
    agent.clean_up()


# Clones the agent and returns an identical agent.
@pytest.mark.parametrize("observation_space", ["vector_space"])
def test_clone_returns_identical_agent(observation_space, discrete_space, request):
    observation_space = request.getfixturevalue(observation_space)
    ppo = DummyPPO(observation_space, discrete_space)
    ppo.fitness = [200, 200, 200]
    ppo.scores = [94, 94, 94]
    ppo.steps = [2500]
    ppo.tensor_attribute = torch.randn(1)
    clone_agent = ppo.clone()

    assert clone_agent.observation_space == ppo.observation_space
    assert clone_agent.action_space == ppo.action_space
    assert clone_agent.batch_size == ppo.batch_size
    assert clone_agent.lr == ppo.lr
    assert clone_agent.gamma == ppo.gamma
    assert clone_agent.gae_lambda == ppo.gae_lambda
    assert clone_agent.mut == ppo.mut
    assert clone_agent.action_std_init == ppo.action_std_init
    assert clone_agent.clip_coef == ppo.clip_coef
    assert clone_agent.ent_coef == ppo.ent_coef
    assert clone_agent.vf_coef == ppo.vf_coef
    assert clone_agent.max_grad_norm == ppo.max_grad_norm
    assert clone_agent.target_kl == ppo.target_kl
    assert clone_agent.update_epochs == ppo.update_epochs
    assert clone_agent.device == ppo.device
    assert clone_agent.accelerator == ppo.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ppo.actor.state_dict())
    assert_state_dicts_equal(clone_agent.critic.state_dict(), ppo.critic.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), ppo.optimizer.state_dict()
    )
    assert clone_agent.fitness == ppo.fitness
    assert clone_agent.steps == ppo.steps
    assert clone_agent.scores == ppo.scores
    assert clone_agent.tensor_attribute == ppo.tensor_attribute
    assert clone_agent.tensor_test == ppo.tensor_test
    assert clone_agent.num_envs == ppo.num_envs
    assert clone_agent.index == ppo.index

    accelerator = Accelerator()
    ppo = PPO(observation_space, discrete_space, accelerator=accelerator)
    clone_agent = ppo.clone()

    assert clone_agent.observation_space == ppo.observation_space
    assert clone_agent.action_space == ppo.action_space
    assert clone_agent.batch_size == ppo.batch_size
    assert clone_agent.lr == ppo.lr
    assert clone_agent.gamma == ppo.gamma
    assert clone_agent.gae_lambda == ppo.gae_lambda
    assert clone_agent.mut == ppo.mut
    assert clone_agent.action_std_init == ppo.action_std_init
    assert clone_agent.clip_coef == ppo.clip_coef
    assert clone_agent.ent_coef == ppo.ent_coef
    assert clone_agent.vf_coef == ppo.vf_coef
    assert clone_agent.max_grad_norm == ppo.max_grad_norm
    assert clone_agent.target_kl == ppo.target_kl
    assert clone_agent.update_epochs == ppo.update_epochs
    assert clone_agent.device == ppo.device
    assert clone_agent.accelerator == ppo.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ppo.actor.state_dict())
    assert_state_dicts_equal(clone_agent.critic.state_dict(), ppo.critic.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), ppo.optimizer.state_dict()
    )
    assert clone_agent.fitness == ppo.fitness
    assert clone_agent.steps == ppo.steps
    assert clone_agent.scores == ppo.scores
    assert clone_agent.num_envs == ppo.num_envs
    assert clone_agent.index == ppo.index

    accelerator = (
        Accelerator(cpu=True) if torch.backends.mps.is_available() else Accelerator()
    )
    ppo = PPO(
        observation_space,
        discrete_space,
        accelerator=accelerator,
        wrap=False,
    )
    clone_agent = ppo.clone(wrap=False)

    assert clone_agent.observation_space == ppo.observation_space
    assert clone_agent.action_space == ppo.action_space
    assert clone_agent.batch_size == ppo.batch_size
    assert clone_agent.lr == ppo.lr
    assert clone_agent.gamma == ppo.gamma
    assert clone_agent.gae_lambda == ppo.gae_lambda
    assert clone_agent.mut == ppo.mut
    assert clone_agent.action_std_init == ppo.action_std_init
    assert clone_agent.clip_coef == ppo.clip_coef
    assert clone_agent.ent_coef == ppo.ent_coef
    assert clone_agent.vf_coef == ppo.vf_coef
    assert clone_agent.max_grad_norm == ppo.max_grad_norm
    assert clone_agent.target_kl == ppo.target_kl
    assert clone_agent.update_epochs == ppo.update_epochs
    assert clone_agent.device == ppo.device
    assert clone_agent.accelerator == ppo.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ppo.actor.state_dict())
    assert_state_dicts_equal(clone_agent.critic.state_dict(), ppo.critic.state_dict())
    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), ppo.optimizer.state_dict()
    )
    assert clone_agent.fitness == ppo.fitness
    assert clone_agent.steps == ppo.steps
    assert clone_agent.scores == ppo.scores
    assert clone_agent.num_envs == ppo.num_envs


def test_clone_new_index(vector_space, discrete_space):
    ppo = PPO(vector_space, discrete_space)
    clone_agent = ppo.clone(index=100)
    assert clone_agent.index == 100
    ppo.clean_up()
    clone_agent.clean_up()


@pytest.mark.parametrize("device", ["cpu", "cuda"], ids=lambda d: f"device={d}")
@pytest.mark.parametrize(
    "use_rollout_buffer", [True, False], ids=lambda b: f"use_rollout_buffer={b}"
)
@pytest.mark.parametrize("recurrent", [True, False], ids=lambda r: f"recurrent={r}")
@pytest.mark.parametrize(
    "share_encoders", [True, False], ids=lambda s: f"share_encoders={s}"
)
def test_clone_after_learning(
    device, use_rollout_buffer, recurrent, share_encoders, vector_space
):
    # accept if recurrent and no rollout buffer
    if recurrent and not use_rollout_buffer:
        return

    # check if device is available
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA not available")

    observation_space = vector_space
    action_space = vector_space
    max_env_steps = 20
    num_vec_envs = 2

    if recurrent:
        net_config = {
            "encoder_config": {
                "hidden_state_size": 64,
            }
        }
    else:
        net_config = {}

    ppo = PPO(
        observation_space,
        action_space,
        device=torch.device(device),
        use_rollout_buffer=use_rollout_buffer,
        recurrent=recurrent,
        net_config=net_config,
        num_envs=num_vec_envs,
        share_encoders=share_encoders,
        max_seq_len=None,
    )

    if use_rollout_buffer:
        dummy_env = DummyEnv(observation_space.shape, vect=True, num_envs=num_vec_envs)
        # Use the correct rollout collection function based on whether the policy is recurrent
        collect_function = collect_rollouts_recurrent if recurrent else collect_rollouts
        collect_function(ppo, dummy_env)
        ppo.learn()
    else:
        states = np.random.randn(
            max_env_steps, num_vec_envs, observation_space.shape[0]
        )
        next_states = np.random.randn(num_vec_envs, observation_space.shape[0])
        actions = np.random.rand(max_env_steps, num_vec_envs, action_space.shape[0])
        log_probs = -np.random.rand(max_env_steps, num_vec_envs)
        rewards = np.random.randint(0, 100, (max_env_steps, num_vec_envs))
        dones = np.zeros((max_env_steps, num_vec_envs))
        values = np.random.randn(
            max_env_steps,
            num_vec_envs,
        )
        next_done = np.zeros((1, num_vec_envs))
        experiences = (
            states,
            actions,
            log_probs,
            rewards,
            dones,
            values,
            next_states,
            next_done,
        )

        ppo.learn(experiences)

    clone_agent = ppo.clone()

    assert clone_agent.observation_space == ppo.observation_space
    assert clone_agent.action_space == ppo.action_space
    assert clone_agent.batch_size == ppo.batch_size
    assert clone_agent.lr == ppo.lr
    assert clone_agent.gamma == ppo.gamma
    assert clone_agent.gae_lambda == ppo.gae_lambda
    assert clone_agent.mut == ppo.mut
    assert clone_agent.action_std_init == ppo.action_std_init
    assert clone_agent.clip_coef == ppo.clip_coef
    assert clone_agent.ent_coef == ppo.ent_coef
    assert clone_agent.vf_coef == ppo.vf_coef
    assert clone_agent.max_grad_norm == ppo.max_grad_norm
    assert clone_agent.target_kl == ppo.target_kl
    assert clone_agent.update_epochs == ppo.update_epochs
    assert clone_agent.device == ppo.device
    assert clone_agent.accelerator == ppo.accelerator
    assert_state_dicts_equal(clone_agent.actor.state_dict(), ppo.actor.state_dict())

    if share_encoders and recurrent:
        # the critic might be different if share_encoders is True
        # (the encoder state might be different because of the logic in the share_encoder_parameters)
        # The important thing is that the head_net is the same as the encoder is neither ran during
        # the forward of the exploration, nor the learning step.
        assert_state_dicts_equal(
            clone_agent.critic.head_net.state_dict(), ppo.critic.head_net.state_dict()
        )
    else:
        assert_state_dicts_equal(
            clone_agent.critic.state_dict(), ppo.critic.state_dict()
        )

    assert_state_dicts_equal(
        clone_agent.optimizer.state_dict(), ppo.optimizer.state_dict()
    )
    assert clone_agent.fitness == ppo.fitness
    assert clone_agent.steps == ppo.steps
    assert clone_agent.scores == ppo.scores
    assert clone_agent.num_envs == ppo.num_envs
    assert clone_agent.index == ppo.index


# Test PPO initialization with rollout buffer
@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "multidiscrete_space",
        "multibinary_space",
        "image_space",
        "dict_space",
    ],
)
@pytest.mark.parametrize(
    "action_space",
    [
        "discrete_space",
        "vector_space",
        "multidiscrete_space",
        "multibinary_space",
    ],
)
def test_ppo_with_rollout_buffer(observation_space, action_space, request):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        use_rollout_buffer=True,
        learn_step=100,
    )

    assert ppo.use_rollout_buffer
    assert hasattr(ppo, "rollout_buffer")
    assert isinstance(ppo.rollout_buffer, RolloutBuffer)
    assert ppo.rollout_buffer.capacity == ppo.learn_step
    assert not ppo.rollout_buffer.recurrent

    if not isinstance(
        observation_space,
        (spaces.Box, spaces.Discrete, spaces.MultiDiscrete, spaces.MultiBinary),
    ):
        pytest.skip("Recurrent PPO with non-vector space is not supported yet!")

    # Build an encoder configuration that matches the observation space type
    if len(observation_space.shape) == 3:  # Image observations – use CNN
        base_net_config = {
            "encoder_config": {
                "channel_size": [16, 32],
                "kernel_size": [3, 3],
                "stride_size": [1, 1],
            }
        }
        expected_shared = {}
        expected_separate = {}
    else:  # Vector observations – use LSTM
        base_net_config = {
            "encoder_config": {
                "hidden_state_size": 64,
            }
        }
        expected_shared = {
            "shared_encoder_h": (1, 1, 64),
            "shared_encoder_c": (1, 1, 64),
        }
        expected_separate = {
            "actor_encoder_h": (1, 1, 64),
            "actor_encoder_c": (1, 1, 64),
            "critic_encoder_h": (1, 1, 64),
            "critic_encoder_c": (1, 1, 64),
        }

    # Recurrent only when hidden states are expected (vector observations)
    recurrent_flag = len(expected_shared) > 0
    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        recurrent=recurrent_flag,
        use_rollout_buffer=True,
        max_seq_len=10,
        net_config=base_net_config,
    )

    if recurrent_flag:
        assert ppo.recurrent
        assert ppo.rollout_buffer.recurrent
        assert ppo.rollout_buffer.hidden_state_architecture == expected_shared

        # Test with separated encoders when hidden states exist
        base_net_config_share = base_net_config.copy()
        ppo = PPO(
            observation_space=observation_space,
            action_space=action_space,
            recurrent=True,
            use_rollout_buffer=True,
            share_encoders=False,
            max_seq_len=10,
            net_config=base_net_config_share,
        )

        assert ppo.rollout_buffer.hidden_state_architecture == expected_separate
        assert not ppo.share_encoders

    # Test with hidden states / separated encoders
    if expected_separate:
        base_net_config_share = base_net_config.copy()
        ppo = PPO(
            observation_space=observation_space,
            action_space=action_space,
            recurrent=True,
            use_rollout_buffer=True,
            share_encoders=False,
            max_seq_len=10,
            net_config=base_net_config_share,
        )

        assert ppo.recurrent
        assert ppo.rollout_buffer.hidden_state_architecture == expected_separate
        assert ppo.rollout_buffer.recurrent
        assert not ppo.share_encoders
        ppo.clean_up()


# Test PPO learning with rollout buffer
@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "discrete_space",
        "multidiscrete_space",
        "multibinary_space",
        "image_space",
        "dict_space",
    ],
)
@pytest.mark.parametrize(
    "action_space",
    ["discrete_space", "vector_space", "multidiscrete_space", "multibinary_space"],
)
@pytest.mark.parametrize("recurrent", [True, False])
@pytest.mark.parametrize("max_seq_len", [None, 10])
@pytest.mark.parametrize(
    "bptt_sequence_type",
    [
        BPTTSequenceType.CHUNKED,
        BPTTSequenceType.MAXIMUM,
        BPTTSequenceType.FIFTY_PERCENT_OVERLAP,
    ],
)
def test_ppo_learn_with_rollout_buffer(
    observation_space, action_space, bptt_sequence_type, recurrent, max_seq_len, request
):
    supported_spaces = [
        "vector_space",
        "discrete_space",
        "multidiscrete_space",
        "multibinary_space",
    ]
    if recurrent and observation_space not in supported_spaces:
        pytest.skip("Recurrent PPO with non-vector space is not supported yet!")

    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    batch_size = 32
    learn_step = 64

    if recurrent:
        net_config = {
            "encoder_config": {
                "hidden_state_size": 64,
            }
        }
    else:
        net_config = {}

    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        use_rollout_buffer=True,
        learn_step=learn_step,
        batch_size=batch_size,
        bptt_sequence_type=bptt_sequence_type,
        recurrent=recurrent,
        max_seq_len=max_seq_len,
        net_config=net_config,
    )

    # Fill the buffer manually
    for i in range(learn_step):
        obs, next_obs = get_batch_states(observation_space, 1)
        action = np.array([action_space.sample()], dtype=action_space.dtype)
        reward = 1.0
        done = i == (learn_step - 1)  # Last step is done
        value = 0.5
        log_prob = -0.5
        if recurrent:
            hidden_state = ppo.get_initial_hidden_state()
            ppo.rollout_buffer.add(
                obs, action, reward, done, value, log_prob, next_obs, hidden_state
            )
        else:
            ppo.rollout_buffer.add(obs, action, reward, done, value, log_prob, next_obs)

    # Compute returns and advantages (normally called by collect_rollouts)
    # For manual filling, we might need to call it if not implicitly handled by learn()
    # However, PPO.learn() calls buffer.compute_returns_and_advantages if experiences are None
    # So, this explicit call might not be strictly necessary if learn() is called without experiences.
    # For clarity in testing buffer functionality, it's fine.
    last_value = torch.zeros((ppo.num_envs, 1), device=ppo.device)
    last_done = torch.zeros((ppo.num_envs, 1), device=ppo.device)
    ppo.rollout_buffer.compute_returns_and_advantages(last_value, last_done)

    # Learn from rollout buffer
    loss = ppo.learn()

    assert isinstance(loss, float)
    assert loss >= 0.0
    ppo.clean_up()


# Test PPO with hidden states
@pytest.mark.parametrize("use_rollout_buffer", [True, False])
@pytest.mark.parametrize("max_seq_len", [None, 10])
def test_ppo_with_hidden_states(
    vector_space, discrete_space, use_rollout_buffer, max_seq_len
):
    observation_space = vector_space
    action_space = discrete_space

    net_config = {
        "encoder_config": {
            "hidden_state_size": 64,
        }
    }

    def make_ppo():
        return PPO(
            observation_space=observation_space,
            action_space=action_space,
            use_rollout_buffer=use_rollout_buffer,
            recurrent=True,
            max_seq_len=max_seq_len,
            net_config=net_config,
        )

    if use_rollout_buffer:
        ppo = make_ppo()
    else:
        with pytest.raises(ValueError):
            ppo = make_ppo()
        return

    # Get action with hidden state
    obs = np.random.rand(1, *observation_space.shape).astype(
        observation_space.dtype
    )  # Add batch dim for num_envs=1
    hidden_state = ppo.get_initial_hidden_state()

    action, log_prob, entropy, value, next_hidden = ppo.get_action(
        obs, hidden_state=hidden_state
    )

    assert action.shape[0] == 1
    assert isinstance(log_prob, np.ndarray)
    assert isinstance(entropy, np.ndarray)
    assert isinstance(value, np.ndarray)
    assert next_hidden is not None
    assert next_hidden.get("shared_encoder_h", None).shape == (
        1,
        1,
        64,
    )  # (directions, num_envs, hidden_size)
    assert next_hidden.get("shared_encoder_c", None).shape == (1, 1, 64)
    ppo.clean_up()


# Test PPO with hidden states
def test_ppo_with_hidden_states_multiple_obs(vector_space, discrete_space):
    observation_space = vector_space
    action_space = discrete_space
    num_envs = 2

    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        use_rollout_buffer=True,
        recurrent=True,
        num_envs=num_envs,
        net_config={
            "encoder_config": {
                "hidden_state_size": 64,
            }
        },
    )

    # Get action with hidden state (multiple observations)
    obs = np.zeros((num_envs, *observation_space.shape), dtype=observation_space.dtype)
    hidden_state = ppo.get_initial_hidden_state(num_envs=num_envs)

    action, log_prob, entropy, value, next_hidden = ppo.get_action(
        obs, hidden_state=hidden_state
    )

    assert action.shape[0] == num_envs
    assert isinstance(log_prob, np.ndarray)
    assert isinstance(entropy, np.ndarray)
    assert isinstance(value, np.ndarray)
    assert next_hidden is not None
    assert next_hidden.get("shared_encoder_h", None).shape == (1, num_envs, 64)
    assert next_hidden.get("shared_encoder_c", None).shape == (1, num_envs, 64)
    ppo.clean_up()


# Test PPO with hidden states
def test_ppo_with_hidden_states_multiple_envs():
    num_envs = 2
    env = gymnasium.vector.SyncVectorEnv(
        [lambda: gymnasium.make("CartPole-v1")] * num_envs
    )

    observation_space = env.single_observation_space  # Use single env space
    action_space = env.single_action_space  # Use single env space

    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        use_rollout_buffer=True,
        recurrent=True,
        num_envs=num_envs,
        max_seq_len=10,
        net_config={
            "encoder_config": {
                "hidden_state_size": 64,
            }
        },
    )

    # Get action with hidden state (multiple observations)
    obs, _ = env.reset()
    hidden_state = ppo.get_initial_hidden_state(num_envs=num_envs)

    action, log_prob, entropy, value, next_hidden = ppo.get_action(
        obs, hidden_state=hidden_state
    )

    assert action.shape[0] == num_envs
    assert isinstance(log_prob, np.ndarray)
    assert isinstance(entropy, np.ndarray)
    assert isinstance(value, np.ndarray)
    assert next_hidden is not None
    assert next_hidden.get("shared_encoder_h", None).shape == (1, num_envs, 64)
    assert next_hidden.get("shared_encoder_c", None).shape == (1, num_envs, 64)
    ppo.clean_up()
    env.close()


# Test PPO with hidden states and collect_rollouts
def test_ppo_with_hidden_states_multiple_envs_collect_rollouts():
    num_envs = 2
    env = gymnasium.vector.SyncVectorEnv(
        [lambda: gymnasium.make("CartPole-v1")] * num_envs
    )

    observation_space = env.single_observation_space  # Use single env space
    action_space = env.single_action_space  # Use single env space

    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        use_rollout_buffer=True,
        recurrent=True,
        num_envs=num_envs,
        learn_step=5,
        max_seq_len=5,
        net_config={
            "encoder_config": {
                "hidden_state_size": 64,
            }
        },
    )

    # Collect rollouts with recurrent network
    collect_rollouts_recurrent(ppo, env)

    # Check buffer contents
    assert ppo.rollout_buffer.pos == -(ppo.learn_step // -ppo.num_envs)
    assert ppo.rollout_buffer.recurrent is True
    # Check observation for the first env at the first timestep
    assert not np.array_equal(
        ppo.rollout_buffer.buffer.get("observations")[0, 0].cpu().numpy(),
        np.zeros(observation_space.shape, dtype=observation_space.dtype),
    )
    # Check actions for all envs at the first timestep
    assert ppo.rollout_buffer.buffer.get("actions") is not None

    hidden_states = ppo.rollout_buffer.buffer.get("hidden_states")
    assert hidden_states is not None
    assert hidden_states.get("shared_encoder_h") is not None
    assert hidden_states.get("shared_encoder_c") is not None

    # Verify hidden states were properly stored (first step's hidden state)
    assert hidden_states.get("shared_encoder_h")[0].shape == (
        num_envs,
        1,  # num_layers * directions
        64,  # hidden_size
    )
    assert hidden_states.get("shared_encoder_c")[0].shape == (
        num_envs,
        1,
        64,
    )

    # Learn from collected rollouts
    loss = ppo.learn()

    assert isinstance(loss, float)
    assert loss >= 0.0
    ppo.clean_up()
    env.close()


# Test PPO with hidden states and collect_rollouts
def test_ppo_with_hidden_states_multiple_envs_collect_rollouts_and_test():
    num_envs = 8
    env = gymnasium.vector.SyncVectorEnv(
        [lambda: gymnasium.make("CartPole-v1")] * num_envs
    )
    num_test_envs = 2
    test_env = gymnasium.vector.SyncVectorEnv(
        [lambda: gymnasium.make("CartPole-v1")] * num_test_envs
    )

    observation_space = env.single_observation_space  # Use single env space
    action_space = env.single_action_space  # Use single env space

    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        use_rollout_buffer=True,
        recurrent=True,
        num_envs=num_envs,
        learn_step=5,
        max_seq_len=5,
        net_config={
            "encoder_config": {
                "hidden_state_size": 64,
            }
        },
    )

    # Collect rollouts with recurrent network
    collect_rollouts_recurrent(ppo, env)

    # Check buffer contents
    assert ppo.rollout_buffer.pos == -(ppo.learn_step // -ppo.num_envs)
    assert ppo.rollout_buffer.recurrent is True
    assert not np.array_equal(
        ppo.rollout_buffer.buffer.get("observations")[0, 0].cpu().numpy(),
        np.zeros(observation_space.shape, dtype=observation_space.dtype),
    )
    assert ppo.rollout_buffer.buffer.get("actions")[0].cpu().numpy() is not None

    assert ppo.rollout_buffer.buffer.get("hidden_states") is not None

    # Verify hidden states were properly stored
    assert ppo.rollout_buffer.buffer.get("hidden_states").get("shared_encoder_h")[
        0
    ].shape == (
        num_envs,
        1,
        64,
    )
    assert ppo.rollout_buffer.buffer.get("hidden_states").get("shared_encoder_c")[
        0
    ].shape == (
        num_envs,
        1,
        64,
    )

    # Learn from collected rollouts
    loss = ppo.learn()

    assert isinstance(loss, float)
    assert loss >= 0.0

    # Test test loop
    ppo.test(test_env)
    ppo.clean_up()
    env.close()
    test_env.close()


# Test PPO collect_rollouts method
@pytest.mark.parametrize(
    "observation_space",
    [
        "vector_space",
        "image_space",
        # "dict_space",
    ],
)
@pytest.mark.parametrize(
    "action_space",
    [
        "discrete_space",
        "vector_space",
        "multidiscrete_space",
        "multibinary_space",
    ],
)
@pytest.mark.parametrize(
    "bptt_sequence_type",
    [
        BPTTSequenceType.CHUNKED,
        BPTTSequenceType.MAXIMUM,
        BPTTSequenceType.FIFTY_PERCENT_OVERLAP,
    ],
)
def test_ppo_collect_rollouts(
    observation_space, action_space, bptt_sequence_type, request
):
    observation_space = request.getfixturevalue(observation_space)
    action_space = request.getfixturevalue(action_space)

    learn_step = 5

    ppo = PPO(
        observation_space=observation_space,
        action_space=action_space,
        use_rollout_buffer=True,
        learn_step=learn_step,
        num_envs=1,  # Explicitly set num_envs for clarity
        bptt_sequence_type=bptt_sequence_type,
    )

    env = DummyEnv(state_size=observation_space.shape, vect=True, num_envs=ppo.num_envs)

    # Collect rollouts
    collect_rollouts(ppo, env, n_steps=learn_step)

    # Check if properties and weights are loaded correctly
    assert ppo.observation_space == ppo.observation_space
    assert ppo.action_space == ppo.action_space
    assert isinstance(ppo.actor, nn.Module)
    assert isinstance(ppo.critic, nn.Module)
    assert ppo.lr == ppo.lr
    assert str(ppo.actor.to("cpu").state_dict()) == str(ppo.actor.state_dict())
    assert str(ppo.critic.to("cpu").state_dict()) == str(ppo.critic.state_dict())
    assert ppo.batch_size == ppo.batch_size
    assert ppo.gamma == ppo.gamma
    assert ppo.mut == ppo.mut
    assert ppo.index == ppo.index
    assert ppo.scores == ppo.scores
    assert ppo.fitness == ppo.fitness
    assert ppo.steps == ppo.steps
    ppo.clean_up()



import gc

import numpy as np
import pytest
import torch

from agilerl.algorithms.ippo import IPPO
from agilerl.algorithms.maddpg import MADDPG
from agilerl.components.multi_agent_replay_buffer import MultiAgentReplayBuffer
from agilerl.utils.probe_envs_ma import (
    ConstantRewardContActionsEnv,
    ConstantRewardContActionsImageEnv,
    ConstantRewardEnv,
    ConstantRewardImageEnv,
    DiscountedRewardContActionsEnv,
    DiscountedRewardContActionsImageEnv,
    DiscountedRewardEnv,
    DiscountedRewardImageEnv,
    FixedObsPolicyContActionsEnv,
    FixedObsPolicyContActionsImageEnv,
    FixedObsPolicyEnv,
    FixedObsPolicyImageEnv,
    MultiPolicyEnv,
    MultiPolicyImageEnv,
    ObsDependentRewardContActionsEnv,
    ObsDependentRewardContActionsImageEnv,
    ObsDependentRewardEnv,
    ObsDependentRewardImageEnv,
    PolicyContActionsEnv,
    PolicyContActionsImageEnv,
    PolicyEnv,
    PolicyImageEnv,
    check_on_policy_learning_with_probe_env,
    check_policy_q_learning_with_probe_env,
)


@pytest.mark.parametrize(
    "env_class, exp_states, exp_rewards, exp_terminateds, exp_truncateds, exp_infos",
    [
        (ConstantRewardEnv, [0, 0], [1, 0], [True, True], [False, False], {}),
        (ConstantRewardImageEnv, [0, 0], [1, 0], [True, True], [False, False], {}),
        (
            ConstantRewardContActionsEnv,
            [0, 0],
            [1, 0],
            [True, True],
            [False, False],
            {},
        ),
        (
            ConstantRewardContActionsImageEnv,
            [0, 0],
            [1, 0],
            [True, True],
            [False, False],
            {},
        ),
    ],
)
def test_constant_reward_envs(
    env_class, exp_states, exp_rewards, exp_terminateds, exp_truncateds, exp_infos
):
    env = env_class()
    states, infos = env.reset()
    for _ in range(20):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            reward,
            terminated,
            truncated,
            info,
            exp_state,
            exp_reward,
            exp_terminated,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            exp_states,
            exp_rewards,
            exp_terminateds,
            exp_truncateds,
            exp_infos,
        ):

            assert int(np.mean(np.array(state))) == exp_state
            assert reward == exp_reward
            assert terminated == exp_terminated
            assert truncated == exp_truncated
            assert info == exp_info

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_states, exp_rewards, exp_terminateds, exp_truncateds, exp_infos",
    [
        (ObsDependentRewardEnv, [0, 0], [1, 0], [True, True], [False, False], {}),
        (ObsDependentRewardEnv, [1, 1], [0, 1], [True, True], [False, False], {}),
        (ObsDependentRewardImageEnv, [0, 0], [1, 0], [True, True], [False, False], {}),
        (ObsDependentRewardImageEnv, [1, 1], [0, 1], [True, True], [False, False], {}),
        (
            ObsDependentRewardContActionsEnv,
            [0, 0],
            [1, 0],
            [True, True],
            [False, False],
            {},
        ),
        (
            ObsDependentRewardContActionsEnv,
            [1, 1],
            [0, 1],
            [True, True],
            [False, False],
            {},
        ),
        (
            ObsDependentRewardContActionsImageEnv,
            [0, 0],
            [1, 0],
            [True, True],
            [False, False],
            {},
        ),
        (
            ObsDependentRewardContActionsImageEnv,
            [1, 1],
            [0, 1],
            [True, True],
            [False, False],
            {},
        ),
    ],
)
def test_observation_dependent_reward_envs(
    env_class, exp_states, exp_rewards, exp_terminateds, exp_truncateds, exp_infos
):
    env = env_class()
    states, info = env.reset()
    for _ in range(20):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            reward,
            terminated,
            truncated,
            info,
            exp_state,
            exp_reward,
            exp_terminated,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            exp_states,
            exp_rewards,
            exp_terminateds,
            exp_truncateds,
            exp_infos,
        ):
            if int(np.mean(np.array(state))) == exp_state:
                assert reward == exp_reward
            assert terminated == exp_terminated
            assert truncated == exp_truncated
            assert info == exp_info

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_states, exp_rewards, exp_truncateds, exp_infos",
    [
        (DiscountedRewardEnv, [0, 0], [0, 0], [False, False], {}),
        (DiscountedRewardEnv, [1, 1], [1, 0.5], [False, False], {}),
        (DiscountedRewardImageEnv, [0, 0], [0, 0], [False, False], {}),
        (DiscountedRewardImageEnv, [1, 1], [1, 0.5], [False, False], {}),
        (DiscountedRewardContActionsEnv, [0, 0], [0, 0], [False, False], {}),
        (DiscountedRewardContActionsEnv, [1, 1], [1, 0.5], [False, False], {}),
        (DiscountedRewardContActionsImageEnv, [0, 0], [0, 0], [False, False], {}),
        (DiscountedRewardContActionsImageEnv, [1, 1], [1, 0.5], [False, False], {}),
    ],
)
def test_discounted_reward_envs(
    env_class, exp_states, exp_rewards, exp_truncateds, exp_infos
):
    env = env_class()
    states, info = env.reset()
    for _ in range(20):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        next_states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            reward,
            terminated,
            truncated,
            info,
            exp_state,
            exp_reward,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            exp_states,
            exp_rewards,
            exp_truncateds,
            exp_infos,
        ):
            if int(np.mean(np.array(state))) == exp_state:
                assert reward == exp_reward
            assert terminated == int(np.mean(np.array(state)))
            assert truncated == exp_truncated
            assert info == exp_info

        states = next_states

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_states, exp_actions, exp_rewards, exp_terminateds, exp_truncateds, exp_infos",
    [
        (FixedObsPolicyEnv, [0, 0], [0, 0], [1, -1], [True, True], [False, False], {}),
        (FixedObsPolicyEnv, [0, 0], [1, 1], [-1, 1], [True, True], [False, False], {}),
        (
            FixedObsPolicyImageEnv,
            [0, 0],
            [0, 0],
            [1, -1],
            [True, True],
            [False, False],
            {},
        ),
        (
            FixedObsPolicyImageEnv,
            [0, 0],
            [1, 1],
            [-1, 1],
            [True, True],
            [False, False],
            {},
        ),
    ],
)
def test_discrete_actions_fixed_observation_policy_reward_envs(
    env_class,
    exp_states,
    exp_actions,
    exp_rewards,
    exp_terminateds,
    exp_truncateds,
    exp_infos,
):
    env = env_class()
    states, info = env.reset()
    for _ in range(20):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            action,
            reward,
            terminated,
            truncated,
            info,
            exp_state,
            exp_action,
            exp_reward,
            exp_terminated,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(actions.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            exp_states,
            exp_actions,
            exp_rewards,
            exp_terminateds,
            exp_truncateds,
            exp_infos,
        ):
            assert int(np.mean(np.array(state))) == exp_state
            if action == exp_action:
                assert reward == exp_reward
            assert terminated == exp_terminated
            assert truncated == exp_truncated
            assert info == exp_info

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


@pytest.mark.parametrize(
    "env_class, exp_states, exp_actions, exp_terminateds, exp_truncateds, exp_infos",
    [
        (
            FixedObsPolicyContActionsEnv,
            [0, 0],
            [1, 0],
            [True, True],
            [False, False],
            {},
        ),
        (
            FixedObsPolicyContActionsImageEnv,
            [0, 0],
            [1, 0],
            [True, True],
            [False, False],
            {},
        ),
    ],
)
def test_continuous_actions_fixed_observation_policy_reward_envs(
    env_class, exp_states, exp_actions, exp_terminateds, exp_truncateds, exp_infos
):
    env = env_class()
    states, info = env.reset()
    for _ in range(20):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            action,
            reward,
            terminated,
            truncated,
            info,
            exp_state,
            exp_action,
            exp_terminated,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(actions.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            exp_states,
            exp_actions,
            exp_terminateds,
            exp_truncateds,
            exp_infos,
        ):
            assert int(np.mean(np.array(state))) == exp_state
            assert reward == -((exp_action - action[0]) ** 2)
            assert terminated == exp_terminated
            assert truncated == exp_truncated
            assert info == exp_info

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


@pytest.mark.parametrize(
    "env_class, same_rewards, diff_rewards, exp_terminateds, exp_truncateds, exp_infos",
    [
        (PolicyEnv, [1, 0], [0, 1], [True, True], [False, False], {}),
        (PolicyImageEnv, [1, 0], [0, 1], [True, True], [False, False], {}),
    ],
)
def test_discrete_actions_policy_envs(
    env_class, same_rewards, diff_rewards, exp_terminateds, exp_truncateds, exp_infos
):
    env = env_class()
    states, info = env.reset()
    for _ in range(20):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            action,
            reward,
            terminated,
            truncated,
            info,
            same_reward,
            diff_reward,
            exp_terminated,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(actions.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            same_rewards,
            diff_rewards,
            exp_terminateds,
            exp_truncateds,
            exp_infos,
        ):
            if int(np.mean(np.array(state))) == action:
                assert reward == same_reward
            else:
                assert reward == diff_reward
            assert terminated == exp_terminated
            assert truncated == exp_truncated
            assert info == exp_info

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


@pytest.mark.parametrize(
    "env_class, reward_goal_0s, reward_goal_1s, exp_terminateds, exp_truncateds, exp_infos",
    [
        (PolicyContActionsEnv, [0, 1], [1, 0], [True, True], [False, False], {}),
        (PolicyContActionsImageEnv, [0, 1], [1, 0], [True, True], [False, False], {}),
    ],
)
def test_continuous_actions_policy_envs(
    env_class,
    reward_goal_0s,
    reward_goal_1s,
    exp_terminateds,
    exp_truncateds,
    exp_infos,
):
    env = env_class()
    states, info = env.reset()
    for _ in range(60):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            action,
            reward,
            terminated,
            truncated,
            info,
            reward_goal_0,
            reward_goal_1,
            exp_terminated,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(actions.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            reward_goal_0s,
            reward_goal_1s,
            exp_terminateds,
            exp_truncateds,
            exp_infos,
        ):
            if int(np.mean(np.array(state))):
                assert (
                    reward
                    == -((reward_goal_0 - action[0]) ** 2)
                    - (reward_goal_1 - action[1]) ** 2
                )
            else:
                assert (
                    reward
                    == -((reward_goal_1 - action[0]) ** 2)
                    - (reward_goal_0 - action[1]) ** 2
                )
            assert terminated == exp_terminated
            assert truncated == exp_truncated
            assert info == exp_info

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


@pytest.mark.parametrize(
    "env_class, same_rewards, diff_rewards, exp_terminateds, exp_truncateds, exp_infos",
    [
        (MultiPolicyEnv, [1, 0], [0, 1], [True, True], [False, False], {}),
        (MultiPolicyImageEnv, [1, 0], [0, 1], [True, True], [False, False], {}),
    ],
)
def test_discrete_actions_multi_policy_envs(
    env_class, same_rewards, diff_rewards, exp_terminateds, exp_truncateds, exp_infos
):
    env = env_class()
    states, info = env.reset()
    for _ in range(20):
        actions = {
            env.agents[0]: env.action_space[env.agents[0]].sample(),
            env.agents[1]: env.action_space[env.agents[1]].sample(),
        }
        states, rewards, terminateds, truncateds, infos = env.step(actions)

        for (
            state,
            action,
            reward,
            terminated,
            truncated,
            info,
            same_reward,
            diff_reward,
            exp_terminated,
            exp_truncated,
            exp_info,
        ) in zip(
            list(states.values()),
            list(actions.values()),
            list(rewards.values()),
            list(terminateds.values()),
            list(truncateds.values()),
            list(infos.values()),
            same_rewards,
            diff_rewards,
            exp_terminateds,
            exp_truncateds,
            exp_infos,
        ):
            if int(np.mean(np.array(state))) == action:
                assert reward == same_reward
            else:
                assert reward == diff_reward
            assert terminated == exp_terminated
            assert truncated == exp_truncated
            assert info == exp_info

        if terminateds[env.agents[0]] or truncateds[env.agents[0]]:
            states, infos = env.reset()


def test_policy_q_learning_with_probe_env():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = ConstantRewardEnv()
    learn_steps = 10
    algo_args = {
        "observation_spaces": [env.observation_space[agent] for agent in env.agents],
        "action_spaces": [env.action_space[agent] for agent in env.agents],
        "agent_ids": env.possible_agents,
        "net_config": {"encoder_config": {"hidden_size": [32, 32]}},
        "batch_size": 256,
    }
    field_names = ["state", "action", "reward", "next_state", "done"]
    memory = MultiAgentReplayBuffer(
        memory_size=10000,  # Max replay buffer size
        field_names=field_names,  # Field names to store in memory
        agent_ids=algo_args["agent_ids"],
        device=device,
    )

    check_policy_q_learning_with_probe_env(
        env, MADDPG, algo_args, memory, learn_steps, device
    )


def test_policy_q_learning_with_probe_env_mlp():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = FixedObsPolicyContActionsEnv()
    learn_steps = 10
    algo_args = {
        "observation_spaces": [env.observation_space[agent] for agent in env.agents],
        "action_spaces": [env.action_space[agent] for agent in env.agents],
        "agent_ids": env.possible_agents,
        "net_config": {"encoder_config": {"hidden_size": [32, 32]}},
        "batch_size": 256,
    }
    field_names = ["state", "action", "reward", "next_state", "done"]
    memory = MultiAgentReplayBuffer(
        memory_size=10000,  # Max replay buffer size
        field_names=field_names,  # Field names to store in memory
        agent_ids=algo_args["agent_ids"],
        device=device,
    )

    check_policy_q_learning_with_probe_env(
        env, MADDPG, algo_args, memory, learn_steps, device
    )
    gc.collect()


def test_policy_q_learning_with_probe_env_cnn():
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = FixedObsPolicyContActionsImageEnv()
    learn_steps = 10
    algo_args = {
        "observation_spaces": [env.observation_space[agent] for agent in env.agents],
        "action_spaces": [env.action_space[agent] for agent in env.agents],
        "agent_ids": env.possible_agents,
        "net_config": {
            "encoder_config": {
                "channel_size": [16],
                "kernel_size": [3],
                "stride_size": [1],
                "init_layers": True,
            },
            "head_config": {
                "hidden_size": [32],
                "init_layers": True,
            },
        },
        "lr_actor": 1e-5,  # Reduced actor learning rate
        "lr_critic": 1e-4,  # Reduced critic learning rate
        "batch_size": 64,  # Smaller batch size
        "normalize_images": True,  # Ensure image normalization
        "gamma": 0.99,  # Stable discount factor
        "tau": 0.005,  # Smaller soft update parameter
    }
    field_names = ["state", "action", "reward", "next_state", "done"]
    memory = MultiAgentReplayBuffer(
        memory_size=10000,
        field_names=field_names,
        agent_ids=algo_args["agent_ids"],
        device=device,
    )

    check_policy_q_learning_with_probe_env(
        env, MADDPG, algo_args, memory, learn_steps, device
    )
    gc.collect()


@pytest.mark.parametrize(
    "Env, discrete",
    [
        (ConstantRewardEnv, True),
        (ConstantRewardContActionsEnv, False),
        (FixedObsPolicyEnv, True),
        (FixedObsPolicyContActionsEnv, False),
    ],
)
def test_on_policy_learning_with_probe_env_mlp(Env, discrete):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = Env()
    learn_steps = 20
    algo_args = {
        "observation_spaces": [space for space in env.observation_space.values()],
        "action_spaces": [space for space in env.action_space.values()],
        "agent_ids": env.agents,
        "lr": 1e-2,
        "net_config": {
            "encoder_config": {"hidden_size": [16, 16], "init_layers": False},
            "head_config": {"hidden_size": [16], "init_layers": False},
        },
    }

    check_on_policy_learning_with_probe_env(
        env, IPPO, algo_args, learn_steps, device, discrete=discrete
    )
    gc.collect()


@pytest.mark.parametrize(
    "Env, discrete",
    [
        (ConstantRewardImageEnv, True),
        (FixedObsPolicyImageEnv, True),
    ],
)
def test_on_policy_learning_with_probe_env_cnn(Env, discrete):
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    env = Env()
    learn_steps = 30
    algo_args = {
        "observation_spaces": [space for space in env.observation_space.values()],
        "action_spaces": [space for space in env.action_space.values()],
        "agent_ids": env.agents,
        "lr": 1e-2,
        "net_config": {
            "encoder_config": {
                "channel_size": [16],
                "kernel_size": [3],
                "stride_size": [1],
            },
            "head_config": {"hidden_size": [32], "output_activation": "Sigmoid"},
        },
        "normalize_images": False,
    }

    check_on_policy_learning_with_probe_env(
        env, IPPO, algo_args, learn_steps, device, discrete=discrete
    )
    gc.collect()
    torch.cuda.empty_cache()



import os

import numpy as np
from accelerate import Accelerator
from tensordict import TensorDictBase
from torch.utils.data import DataLoader

from agilerl.algorithms import DQN
from agilerl.components.data import ReplayDataset, Transition
from agilerl.components.replay_buffer import ReplayBuffer
from agilerl.components.sampler import Sampler
from agilerl.hpo.mutation import Mutations
from agilerl.hpo.tournament import TournamentSelection
from agilerl.utils.utils import (
    create_population,
    default_progress_bar,
    make_vect_envs,
)

# !Note: If you are running this demo without having installed agilerl,
# uncomment and place the following above agilerl imports:

# import sys
# sys.path.append('../')


if __name__ == "__main__":
    accelerator = Accelerator()

    accelerator.wait_for_everyone()
    if accelerator.is_main_process:
        print("===== AgileRL Online Distributed Demo =====")
    accelerator.wait_for_everyone()

    NET_CONFIG = {
        "encoder_config": {
            "hidden_size": [32, 32],  # Actor hidden size
        }
    }

    INIT_HP = {
        "DOUBLE": True,  # Use double Q-learning in DQN or CQN
        "BATCH_SIZE": 128,  # Batch size
        "LR": 1e-3,  # Learning rate
        "GAMMA": 0.99,  # Discount factor
        "LEARN_STEP": 1,  # Learning frequency
        "TAU": 1e-3,  # For soft update of target network parameters
        "POP_SIZE": 4,  # Population size
    }

    num_envs = 8
    env = make_vect_envs("LunarLander-v3", num_envs=num_envs)  # Create environment
    observation_space = env.single_observation_space
    action_space = env.single_action_space

    pop: list[DQN] = create_population(
        algo="DQN",  # Algorithm
        observation_space=observation_space,  # Observation space
        action_space=action_space,  # Action space
        net_config=NET_CONFIG,  # Network configuration
        INIT_HP=INIT_HP,  # Initial hyperparameters
        population_size=INIT_HP["POP_SIZE"],  # Population size
        num_envs=num_envs,  # No. vectorized envs
        accelerator=accelerator,  # Accelerator
    )

    memory = ReplayBuffer(max_size=10000)
    replay_dataset = ReplayDataset(memory, INIT_HP["BATCH_SIZE"])
    replay_dataloader = DataLoader(replay_dataset, batch_size=None)
    replay_dataloader = accelerator.prepare(replay_dataloader)
    sampler = Sampler(dataset=replay_dataset, dataloader=replay_dataloader)

    tournament = TournamentSelection(
        tournament_size=2,  # Tournament selection size
        elitism=True,  # Elitism in tournament selection
        population_size=INIT_HP["POP_SIZE"],  # Population size
        eval_loop=1,  # Evaluate using last N fitness scores
    )

    mutations = Mutations(
        algo="DQN",  # Algorithm
        no_mutation=0.4,  # No mutation
        architecture=0.2,  # Architecture mutation
        new_layer_prob=0.2,  # New layer mutation
        parameters=0.2,  # Network parameters mutation
        activation=0,  # Activation layer mutation
        rl_hp=0.2,  # Learning HP mutation
        mutation_sd=0.1,  # Mutation strength  # Network architecture
        rand_seed=1,  # Random seed
        accelerator=accelerator,
    )  # Accelerator)

    max_steps = 200000  # Max steps
    learning_delay = 1000  # Steps before starting learning

    # Exploration params
    eps_start = 1.0  # Max exploration
    eps_end = 0.1  # Min exploration
    eps_decay = 0.995  # Decay per episode
    epsilon = eps_start

    evo_steps = 10000  # Evolution frequency
    eval_steps = None  # Evaluation steps per episode - go until done
    eval_loop = 1  # Number of evaluation episodes

    total_steps = 0

    accel_temp_models_path = "models/{}".format("LunarLander-v3")
    if accelerator.is_main_process:
        if not os.path.exists(accel_temp_models_path):
            os.makedirs(accel_temp_models_path)

    print(f"\nDistributed training on {accelerator.device}...")

    # TRAINING LOOP
    print("Training...")
    pbar = default_progress_bar(max_steps, accelerator=accelerator)
    while np.less([agent.steps[-1] for agent in pop], max_steps).all():
        accelerator.wait_for_everyone()
        pop_episode_scores = []
        for agent in pop:  # Loop through population
            obs, info = env.reset()  # Reset environment at start of episode
            scores = np.zeros(num_envs)
            completed_episode_scores, losses = [], []
            steps = 0
            epsilon = eps_start

            for idx_step in range(evo_steps):
                # Get next action from agent
                action = agent.get_action(obs, epsilon)
                epsilon = max(
                    eps_end, epsilon * eps_decay
                )  # Decay epsilon for exploration

                # Act in environment
                next_obs, reward, terminated, truncated, info = env.step(action)
                scores += np.array(reward)
                steps += num_envs
                total_steps += num_envs

                # Collect scores for completed episodes
                for idx, (d, t) in enumerate(zip(terminated, truncated)):
                    if d or t:
                        completed_episode_scores.append(scores[idx])
                        agent.scores.append(scores[idx])
                        scores[idx] = 0

                # Save experience to replay buffer
                transition: TensorDictBase = Transition(
                    obs,
                    action,
                    reward,
                    next_obs,
                    terminated,
                )
                transition = transition.to_tensordict()
                transition.batch_size = [num_envs]
                memory.add(transition)

                # Learn according to learning frequency
                if memory.size > learning_delay and len(memory) >= agent.batch_size:
                    for _ in range(num_envs // agent.learn_step):
                        # Sample dataloader
                        experiences = sampler.sample(agent.batch_size)
                        # Learn according to agent's RL algorithm
                        agent.learn(experiences)

                obs = next_obs

            pbar.update(evo_steps // len(pop))
            agent.steps[-1] += steps
            pop_episode_scores.append(completed_episode_scores)

        # Reset epsilon start to latest decayed value for next round of population training
        eps_start = epsilon

        # Evaluate population
        fitnesses = [
            agent.test(
                env,
                max_steps=eval_steps,
                loop=eval_loop,
            )
            for agent in pop
        ]
        mean_scores = [
            (
                np.mean(episode_scores)
                if len(episode_scores) > 0
                else "0 completed episodes"
            )
            for episode_scores in pop_episode_scores
        ]

        if accelerator.is_main_process:
            pbar.write(
                f"--- Global steps {total_steps} ---\n"
                f"Steps: {[agent.steps[-1] for agent in pop]}\n"
                f"Scores: {mean_scores}\n"
                f"Fitnesses: {['%.2f' % fitness for fitness in fitnesses]}\n"
                f"5 fitness avgs: {['%.2f' % np.mean(agent.fitness[-5:]) for agent in pop]}\n"
            )

        # Tournament selection and population mutation
        accelerator.wait_for_everyone()
        for model in pop:
            model.unwrap_models()
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            elite, pop = tournament.select(pop)
            pop = mutations.mutation(pop)
            for pop_i, model in enumerate(pop):
                model.save_checkpoint(f"{accel_temp_models_path}/DQN_{pop_i}.pt")
        accelerator.wait_for_everyone()
        if not accelerator.is_main_process:
            for pop_i, model in enumerate(pop):
                model.load_checkpoint(f"{accel_temp_models_path}/DQN_{pop_i}.pt")
        accelerator.wait_for_everyone()
        for model in pop:
            model.wrap_models()

        # Update step counter
        for agent in pop:
            agent.steps.append(agent.steps[-1])

    pbar.close()
    env.close()
